import os
from threading import Event
from typing import Any, Callable, cast, TYPE_CHECKING

from PyQt6.QtCore import QMutex, Qt, QThread, QTimer, pyqtSignal
from PyQt6.QtWidgets import (
    QHBoxLayout,
    QLayoutItem,
    QSpacerItem,
    QSystemTrayIcon,
    QVBoxLayout,
    QWidget,
)
from senpwai.scrapers import gogo, pahe
from senpwai.utils.classes import SETTINGS, Anime, AnimeDetails
from senpwai.utils.scraper import (
    IBYTES_TO_MBS_DIVISOR,
    Download,
    ProgressFunction,
    ffmpeg_is_installed,
    lacked_episode_numbers,
    lacked_episodes,
    sanitise_title,
)
from senpwai.utils.static import (
    CANCEL_ICON_PATH,
    DOWNLOAD_WINDOW_BCKG_IMAGE_PATH,
    DUB,
    GOGO,
    MOVE_DOWN_QUEUE_ICON_PATH,
    MOVE_UP_QUEUE_ICON_PATH,
    PAHE,
    PAUSE_ICON_PATH,
    REMOVE_FROM_QUEUE_ICON_PATH,
    RESUME_ICON_PATH,
    open_folder,
)
from senpwai.utils.widgets import (
    FolderButton,
    HorizontalLine,
    Icon,
    IconButton,
    OutlinedLabel,
    ProgressBarWithButtons,
    ProgressBarWithoutButtons,
    ScrollableSection,
    StyledButton,
    StyledLabel,
    set_minimum_size_policy,
)

from senpwai.windows.abstracts import AbstractWindow
# https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports/3957388#39757388
if TYPE_CHECKING:
    from senpwai.windows.main import MainWindow


class CurrentAgainstTotal(StyledLabel):
    def __init__(
        self, total: int, units: str, font_size=30, parent: QWidget | None = None
    ):
        super().__init__(parent, font_size)
        self.total = total
        self.current = 0
        self.units = units
        # This is to ensure that even when DownloadedEpisodeCount overwrides update_count, the parent's update count still gets called during parent class initialisation
        CurrentAgainstTotal.update_count(self, 0)

    def update_count(self, added: int):
        self.current += added
        self.setText(f"{self.current}/{self.total} {self.units}")
        set_minimum_size_policy(self)
        self.update()


class HlsEstimatedSize(CurrentAgainstTotal):
    def __init__(self, download_window: 'DownloadWindow', total_episode_count: int):
        super().__init__(0, "MBs", parent=download_window)
        self.total_episode_count = total_episode_count
        self.current_episode_count = 0
        self.sizes_for_each_eps: list[int] = []

    def update_count(self, added: int):
        if added == 0:
            self.total -= self.current
            self.total_episode_count -= 1
        else:
            self.sizes_for_each_eps.append(added)
        count = len(self.sizes_for_each_eps)
        if count > 0:
            new_current = sum(self.sizes_for_each_eps)
            self.current = new_current
            self.total = round((new_current / count) * self.total_episode_count)
        super().update_count(0)


class DownloadedEpisodeCount(CurrentAgainstTotal):
    def __init__(
        self,
        download_window: 'DownloadWindow',
        total_episodes: int,
        anime_title: str,
        anime_folder_path: str,
    ):
        self.download_window = download_window
        self.anime_folder_path = anime_folder_path
        self.anime_title = anime_title
        self.cancelled = False
        super().__init__(total_episodes, "eps", 30, download_window)

    def reinitialise(
        self, new_total: int, new_anime_title: str, new_anime_folder_path: str
    ):
        self.cancelled = False
        self.current = 0
        self.total = new_total
        self.anime_folder_path = new_anime_folder_path
        self.anime_title = new_anime_title
        super().update_count(0)

    def is_complete(self) -> bool:
        return self.current >= self.total

    def update_count(self, added: int):
        super().update_count(added)
        complete = self.is_complete()
        if complete and self.total != 0 and SETTINGS.allow_notifications:
            self.download_window.main_window.tray_icon.make_notification(
                "Download Complete",
                self.anime_title,
                True,
                lambda: open_folder(self.anime_folder_path),
            )
        if complete or self.cancelled:
            self.start_next_download()

    def start_next_download(self):
        queued_downloads_count = len(
            self.download_window.download_queue.get_queued_downloads()
        )
        if queued_downloads_count > 1:
            self.download_window.start_download()


class CancelAllButton(StyledButton):
    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent, 25, "white", "red", "#ED2B2A", "#990000")
        self.setText("CANCEL")
        self.cancel_callback: Callable = lambda: None
        self.clicked.connect(self.cancel)
        self.show()

    def cancel(self):
        self.cancel_callback()


class PauseAllButton(StyledButton):
    def __init__(self, download_is_active: Callable, parent: QWidget | None = None):
        super().__init__(parent, 25, "white", "#FFA41B", "#FFA756", "#F86F03")
        self.setText("PAUSE")
        self.pause_callback: Callable = lambda: None
        self.download_is_active = download_is_active
        self.not_paused_style_sheet = self.styleSheet()
        styles_to_overwride = """
                QPushButton {
                background-color: #FFA756;
            }
            QPushButton:hover {
                background-color: #FFA41B;
            }
            QPushButton:pressed {
                background-color: #F86F03;
            }
            """
        self.paused_style_sheet = self.not_paused_style_sheet + styles_to_overwride
        self.setStyleSheet(self.not_paused_style_sheet)
        self.paused = False
        self.clicked.connect(self.pause_or_resume)
        self.show()

    def pause_or_resume(self):
        if self.download_is_active():
            self.paused = not self.paused
            self.pause_callback()
            if self.paused:
                self.setText("RESUME")
                self.setStyleSheet(self.paused_style_sheet)

            elif not self.paused:
                self.setText("PAUSE")
                self.setStyleSheet(self.not_paused_style_sheet)
            self.update()
            set_minimum_size_policy(self)


class QueuedDownload(QWidget):
    def __init__(
        self,
        anime_details: AnimeDetails,
        progress_bar: ProgressBarWithoutButtons,
        download_queue: 'DownloadQueue',
    ):
        super().__init__()
        label = StyledLabel(font_size=14)
        self.anime_details = anime_details
        self.progress_bar = progress_bar
        label.setText(anime_details.anime.title)
        set_minimum_size_policy(label)
        self.main_layout = QHBoxLayout()
        self.up_button = IconButton(download_queue.up_icon, 1.1, self)
        self.up_button.clicked.connect(
            lambda: download_queue.move_queued_download(self, "up")
        )
        self.down_button = IconButton(download_queue.down_icon, 1.1, self)
        self.down_button.clicked.connect(
            lambda: download_queue.move_queued_download(self, "down")
        )
        self.remove_button = IconButton(download_queue.remove_icon, 1.1, self)
        self.remove_button.clicked.connect(
            lambda: download_queue.remove_queued_download(self)
        )
        self.main_layout.addWidget(label)
        self.main_layout.addWidget(self.up_button)
        self.main_layout.addWidget(self.down_button)
        self.main_layout.addWidget(self.remove_button)
        self.main_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.setLayout(self.main_layout)


class DownloadQueue(QWidget):
    def __init__(self, download_window: QWidget):
        super().__init__(download_window)
        label = OutlinedLabel(None, 1, 25)
        label.setStyleSheet(
            """
            OutlinedLabel {
                color: #4169e1;
                font-size: 25px;
                font-family: "Berlin Sans FB Demi";
                    }
                    """
        )
        label.setText("Download queue")
        main_layout = QVBoxLayout()
        main_layout.addWidget(label)
        self.queued_downloads_layout = QVBoxLayout()
        self.queued_downloads_scrollable = ScrollableSection(
            self.queued_downloads_layout
        )
        line = HorizontalLine(parent=self)
        line.setFixedHeight(6)
        main_layout.addWidget(line)
        main_layout.addWidget(self.queued_downloads_scrollable)
        self.setLayout(main_layout)
        self.up_icon = Icon(30, 30, MOVE_UP_QUEUE_ICON_PATH)
        self.down_icon = Icon(30, 30, MOVE_DOWN_QUEUE_ICON_PATH)
        self.remove_icon = Icon(30, 30, REMOVE_FROM_QUEUE_ICON_PATH)

    def remove_buttons_from_queued_download(self, queued_download: QueuedDownload):
        queued_download.main_layout.removeWidget(queued_download.up_button)
        queued_download.main_layout.removeWidget(queued_download.down_button)
        queued_download.main_layout.removeWidget(queued_download.remove_button)
        queued_download.up_button.deleteLater()
        queued_download.down_button.deleteLater()
        queued_download.remove_button.deleteLater()

    def add_queued_download(
        self, anime_details: AnimeDetails, progress_bar: ProgressBarWithoutButtons
    ):
        self.queued_downloads_layout.addWidget(
            QueuedDownload(anime_details, progress_bar, self),
            alignment=Qt.AlignmentFlag.AlignTop,
        )

    def move_queued_download(self, to_move: QueuedDownload, up_or_down="up"):
        queued_downloads = self.get_queued_downloads()
        for idx, queued in enumerate(queued_downloads):
            if queued == to_move:
                if up_or_down == "up" and idx - 1 > 0:
                    self.queued_downloads_layout.removeWidget(to_move)
                    self.queued_downloads_layout.insertWidget(idx - 1, to_move)
                elif up_or_down == "down" and idx + 1 < len(queued_downloads):
                    self.queued_downloads_layout.removeWidget(to_move)
                    self.queued_downloads_layout.insertWidget(idx + 1, to_move)

    def remove_queued_download(self, queued_download: QueuedDownload):
        for widget in self.get_queued_downloads():
            if widget == queued_download:
                self.queued_downloads_layout.removeWidget(widget)
                widget.deleteLater()

    def get_first_queued_download(self) -> QueuedDownload:
        first_queued_download = cast(
            QueuedDownload,
            cast(QLayoutItem, self.queued_downloads_layout.itemAt(0)).widget(),
        )
        return first_queued_download

    def remove_first_queued_download(self):
        wid = self.get_first_queued_download()
        self.queued_downloads_layout.removeWidget(wid)
        wid.deleteLater()

    def get_queued_downloads(self) -> list[QueuedDownload]:
        count = self.queued_downloads_layout.count()
        return [
            cast(
                QueuedDownload,
                cast(QLayoutItem, self.queued_downloads_layout.itemAt(index)).widget(),
            )
            for index in range(count)
        ]


class DownloadWindow(AbstractWindow):
    def __init__(self, main_window: 'MainWindow'):
        super().__init__(main_window, DOWNLOAD_WINDOW_BCKG_IMAGE_PATH)
        self.main_window = main_window
        self.main_layout = QVBoxLayout()
        self.progress_bars_layout = QVBoxLayout()
        progress_bars_scrollable = ScrollableSection(self.progress_bars_layout)
        top_section_widget = QWidget()
        top_section_layout = QVBoxLayout()
        top_section_widget.setLayout(top_section_layout)
        first_row_of_progress_bar_widget = QWidget()
        self.first_row_of_progress_bar_layout = QHBoxLayout()
        first_row_of_progress_bar_widget.setLayout(
            self.first_row_of_progress_bar_layout
        )
        second_row_of_buttons_widget = QWidget()
        self.second_row_of_buttons_layout = QHBoxLayout()
        second_row_of_buttons_widget.setLayout(self.second_row_of_buttons_layout)
        top_section_layout.addWidget(first_row_of_progress_bar_widget)
        top_section_layout.addWidget(second_row_of_buttons_widget)
        self.main_layout.addWidget(top_section_widget)
        self.main_layout.addWidget(progress_bars_scrollable)
        main_widget = ScrollableSection(self.main_layout)
        self.full_layout.addWidget(main_widget)
        self.setLayout(self.full_layout)
        self.first_download_since_app_start = True
        self.current_anime_progress_bar: ProgressBarWithoutButtons
        self.hls_est_size: HlsEstimatedSize | None = None
        self.download_queue: DownloadQueue
        self.pause_button: PauseAllButton
        self.cancel_button: CancelAllButton
        self.folder_button: FolderButton
        self.downloaded_episode_count: DownloadedEpisodeCount
        self.auto_download_timer = QTimer(self)
        self.auto_download_timer.timeout.connect(self.start_auto_download)
        self.setup_auto_download_timer()
        self.auto_download_thread: AutoDownloadThread | None = None
    
    def is_downloading(self) -> bool:
        if self.first_download_since_app_start:
            return False
        if self.downloaded_episode_count.is_complete() or self.downloaded_episode_count.cancelled:
            return False
        return True

    def setup_auto_download_timer(self):
        self.auto_download_timer.stop()
        self.auto_download_timer.start(  # Converting from hours to milliseconds
            SETTINGS.check_for_new_eps_after * 1000 * 60 * 60
        )

    def clean_out_auto_download_thread(self):
        self.auto_download_thread = None

    def start_auto_download(self):
        # We only spawn a new thread if one wasn't already running to avoid overwriding the reference to the previous one causing it to get garbage collected/destroyed
        # Cause it can cause this error "QThread: Destroyed while thread is still running"
        if SETTINGS.tracked_anime and not self.auto_download_thread:
            self.auto_download_thread = AutoDownloadThread(
                self,
                SETTINGS.tracked_anime,
                self.main_window.tray_icon,
                self.clean_out_auto_download_thread,
            )
            self.auto_download_thread.start()

    def initiate_download_pipeline(self, anime_details: AnimeDetails):
        if self.first_download_since_app_start:
            self.pause_icon = Icon(30, 30, PAUSE_ICON_PATH)
            self.resume_icon = Icon(30, 30, RESUME_ICON_PATH)
            self.cancel_icon = Icon(30, 30, CANCEL_ICON_PATH)
            self.download_queue = DownloadQueue(self)

        if anime_details.sub_or_dub == DUB:
            anime_details.anime.page_link = anime_details.dub_page_link
        if anime_details.site == PAHE:
            return PaheGetEpisodePageInfo(
                self,
                anime_details.lacked_episode_numbers[0],
                anime_details.lacked_episode_numbers[-1],
                anime_details,
                self.pahe_get_episode_page_links,
            ).start()
        self.gogo_get_download_page_links(anime_details)

    def pahe_get_episode_page_links(
        self,
        anime_details: AnimeDetails,
        start_page_num: int,
        end_page_num: int,
        episode_page_count: int,
        first_page: dict[str, Any],
    ):
        episode_page_progress_bar = ProgressBarWithButtons(
            None,
            "Getting episode page links",
            "",
            episode_page_count,
            "pgs",
            1,
            self.pause_icon,
            self.resume_icon,
            self.cancel_icon,
            lambda: None,
            lambda: None,
        )
        self.progress_bars_layout.insertWidget(0, episode_page_progress_bar)
        PaheGetEpisodePageLinksThread(
            self,
            anime_details,
            anime_details.lacked_episode_numbers[0],
            anime_details.lacked_episode_numbers[-1],
            start_page_num,
            end_page_num,
            first_page,
            self.pahe_get_download_page_links,
            episode_page_progress_bar,
        ).start()

    def gogo_get_download_page_links(self, anime_details: AnimeDetails):
        next_func = (
            self.gogo_get_hls_links
            if anime_details.is_hls_download
            else self.get_direct_download_links
        )
        return GogoGetDownloadPageLinksThread(self, anime_details, next_func).start()

    def pahe_get_download_page_links(
        self, anime_details: AnimeDetails, episode_page_links: list[str]
    ):
        episode_page_links = lacked_episodes(
            anime_details.lacked_episode_numbers, episode_page_links
        )
        download_page_progress_bar = ProgressBarWithButtons(
            self,
            "Fetching download page links",
            "",
            len(episode_page_links),
            "eps",
            1,
            self.pause_icon,
            self.resume_icon,
            self.cancel_icon,
            lambda: None,
            lambda: None,
        )
        self.progress_bars_layout.insertWidget(0, download_page_progress_bar)
        PaheGetDownloadPageThread(
            self,
            anime_details,
            episode_page_links,
            self.get_direct_download_links,
            download_page_progress_bar,
        ).start()

    def gogo_get_hls_links(
        self, anime_details: AnimeDetails, episode_page_links: list[str]
    ):
        if not ffmpeg_is_installed():
            return self.main_window.create_and_switch_to_no_ffmpeg_window(anime_details)
        hls_links_progress_bar = ProgressBarWithButtons(
            self,
            "Retrieving hls links, this may take a while",
            "",
            len(episode_page_links),
            "eps",
            1,
            self.pause_icon,
            self.resume_icon,
            self.cancel_icon,
            lambda: None,
            lambda: None,
        )
        self.progress_bars_layout.insertWidget(0, hls_links_progress_bar)
        GetHlsLinksThread(
            self,
            episode_page_links,
            anime_details,
            hls_links_progress_bar,
            self.gogo_get_hls_matched_quality_links,
        ).start()

    def gogo_get_hls_matched_quality_links(
        self, anime_details: AnimeDetails, hls_links: list[str]
    ):
        match_progress_bar = ProgressBarWithButtons(
            self,
            "Matching quality to links",
            "",
            len(hls_links),
            "eps",
            1,
            self.pause_icon,
            self.resume_icon,
            self.cancel_icon,
            lambda: None,
            lambda: None,
        )
        self.progress_bars_layout.insertWidget(0, match_progress_bar)
        GogoHlsGetMatchedQualityLinksThread(
            self,
            hls_links,
            anime_details,
            match_progress_bar,
            self.gogo_hls_get_segments_urls,
        ).start()

    def gogo_hls_get_segments_urls(
        self, anime_details: AnimeDetails, matched_links: list[str]
    ):
        segments_progress_bar = ProgressBarWithButtons(
            self,
            "Getting segment links",
            "",
            len(matched_links),
            "eps",
            1,
            self.pause_icon,
            self.resume_icon,
            self.cancel_icon,
            lambda: None,
            lambda: None,
        )
        self.progress_bars_layout.insertWidget(0, segments_progress_bar)
        GogoHlsGetSegmentsUrlsThread(
            self,
            matched_links,
            anime_details,
            segments_progress_bar,
            self.queue_download,
        ).start()

    def get_direct_download_links(
        self,
        anime_details: AnimeDetails,
        download_page_links: list[str],
        download_info: list[list[str]],
    ):
        direct_download_links_progress_bar = ProgressBarWithButtons(
            self,
            "Retrieving direct download links, this may take a while",
            "",
            len(download_page_links),
            "eps",
            1,
            self.pause_icon,
            self.resume_icon,
            self.cancel_icon,
            lambda: None,
            lambda: None,
        )
        self.progress_bars_layout.insertWidget(0, direct_download_links_progress_bar)
        GetDirectDownloadLinksThread(
            self,
            download_page_links,
            download_info,
            anime_details,
            self.calculate_download_size,
            direct_download_links_progress_bar,
        ).start()

    def calculate_download_size(self, anime_details: AnimeDetails):
        if anime_details.site == PAHE:
            anime_details.total_download_size = pahe.calculate_total_download_size(
                anime_details.download_info
            )
            self.queue_download(anime_details)
        elif anime_details.skip_calculating_size:
            self.queue_download(anime_details)
        else:
            calculating_download_size_progress_bar = ProgressBarWithButtons(
                self,
                "Calculating total download size",
                "",
                len(anime_details.ddls_or_segs_urls),
                "eps",
                1,
                self.pause_icon,
                self.resume_icon,
                self.cancel_icon,
                lambda: None,
                lambda: None,
            )
            self.progress_bars_layout.insertWidget(
                0, calculating_download_size_progress_bar
            )
            GogoCalculateDownloadSizes(
                self,
                anime_details,
                self.queue_download,
                calculating_download_size_progress_bar,
            ).start()

    def queue_download(self, anime_details: AnimeDetails):
        anime_details.validate_anime_folder_path()
        if anime_details.is_hls_download:
            total_segments = sum(
                len(link_or_segs_urls)
                for link_or_segs_urls in anime_details.ddls_or_segs_urls
            )
            anime_progress_bar = ProgressBarWithoutButtons(
                self,
                "Downloading[HLS]",
                anime_details.anime.title,
                total_segments,
                "segs",
                1,
                False,
            )
        elif anime_details.skip_calculating_size:
            anime_progress_bar = ProgressBarWithoutButtons(
                self,
                "Downloading",
                anime_details.anime.title,
                len(anime_details.ddls_or_segs_urls),
                "eps",
                1,
                False,
            )
        else:
            anime_progress_bar = ProgressBarWithoutButtons(
                self,
                "Downloading",
                anime_details.anime.title,
                anime_details.total_download_size,
                "MB",
                1,
                False,
            )
        anime_progress_bar.bar.setMinimumHeight(50)

        self.download_queue.add_queued_download(anime_details, anime_progress_bar)
        if self.first_download_since_app_start:
            self.downloaded_episode_count = DownloadedEpisodeCount(
                self, 0, anime_details.sanitised_title, anime_details.anime_folder_path
            )

            set_minimum_size_policy(self.downloaded_episode_count)
            self.folder_button = FolderButton("", 100, 100, None)

            def download_is_active() -> bool:
                return not (
                    self.downloaded_episode_count.is_complete()
                    or self.downloaded_episode_count.cancelled
                )

            self.pause_button = PauseAllButton(download_is_active, self)
            self.cancel_button = CancelAllButton(self)
            set_minimum_size_policy(self.pause_button)
            set_minimum_size_policy(self.cancel_button)
            self.second_row_of_buttons_layout.addWidget(self.downloaded_episode_count)
            self.second_row_of_buttons_layout.addSpacerItem(QSpacerItem(50, 0))
            self.second_row_of_buttons_layout.addWidget(self.pause_button)
            self.second_row_of_buttons_layout.addWidget(self.cancel_button)
            self.second_row_of_buttons_layout.addSpacerItem(QSpacerItem(50, 0))
            self.second_row_of_buttons_layout.addWidget(self.folder_button)
            self.second_row_of_buttons_layout.addSpacerItem(QSpacerItem(50, 0))
            self.second_row_of_buttons_layout.addWidget(self.download_queue)
        if not self.pause_button.download_is_active():
            self.start_download()

    def clean_out_previous_download(self):
        if self.first_download_since_app_start:
            self.first_download_since_app_start = False
            return
        self.download_queue.remove_first_queued_download()
        self.first_row_of_progress_bar_layout.removeWidget(
            self.current_anime_progress_bar
        )
        self.current_anime_progress_bar.deleteLater()
        if self.hls_est_size:
            self.hls_est_size.deleteLater()
            self.hls_est_size = None

    def start_download(self):
        self.clean_out_previous_download()
        current_queued = self.download_queue.get_first_queued_download()
        self.download_queue.remove_buttons_from_queued_download(current_queued)
        anime_details = current_queued.anime_details
        is_hls_download = anime_details.is_hls_download
        if is_hls_download:
            self.hls_est_size = HlsEstimatedSize(
                self, len(anime_details.ddls_or_segs_urls)
            )
            self.second_row_of_buttons_layout.insertWidget(0, self.hls_est_size)
        self.current_anime_progress_bar = current_queued.progress_bar
        self.downloaded_episode_count.reinitialise(
            len(anime_details.ddls_or_segs_urls),
            anime_details.sanitised_title,
            cast(str, anime_details.anime_folder_path),
        )
        self.first_row_of_progress_bar_layout.addWidget(self.current_anime_progress_bar)
        current_download_manager_thread = DownloadManagerThread(
            self,
            anime_details,
            self.current_anime_progress_bar,
            self.downloaded_episode_count,
        )
        self.pause_button.pause_callback = (
            current_download_manager_thread.pause_or_resume
        )
        self.cancel_button.cancel_callback = current_download_manager_thread.cancel
        self.folder_button.folder_path = anime_details.anime_folder_path
        current_download_manager_thread.start()

    def make_episode_progress_bar(
        self,
        episode_title: str,
        episode_size_or_segs: int,
        progress_bars: dict[str, ProgressBarWithButtons],
        is_hls_download: bool,
    ):
        if is_hls_download:
            bar = ProgressBarWithButtons(
                None,
                "Downloading[HLS]",
                episode_title,
                episode_size_or_segs,
                "segs",
                1,
                self.pause_icon,
                self.resume_icon,
                self.cancel_icon,
                lambda: None,
                lambda: None,
            )
        else:
            bar = ProgressBarWithButtons(
                None,
                "Downloading",
                episode_title,
                episode_size_or_segs,
                "MB",
                IBYTES_TO_MBS_DIVISOR,
                self.pause_icon,
                self.resume_icon,
                self.cancel_icon,
                lambda: None,
                lambda: None,
            )
        progress_bars[episode_title] = bar
        self.progress_bars_layout.insertWidget(0, bar)


class DownloadManagerThread(QThread, ProgressFunction):
    send_progress_bar_details = pyqtSignal(str, int, dict, bool)
    update_anime_progress_bar_signal = pyqtSignal(int)

    def __init__(
        self,
        download_window: DownloadWindow,
        anime_details: AnimeDetails,
        anime_progress_bar: ProgressBarWithoutButtons,
        downloaded_episode_count: DownloadedEpisodeCount,
    ) -> None:
        QThread.__init__(self, download_window)
        ProgressFunction.__init__(self)
        self.anime_progress_bar = anime_progress_bar
        self.download_window = download_window
        self.downloaded_episode_count = downloaded_episode_count
        self.anime_details = anime_details
        self.update_anime_progress_bar_signal.connect(anime_progress_bar.update_bar)
        self.send_progress_bar_details.connect(
            download_window.make_episode_progress_bar
        )
        self.progress_bars: dict[str, ProgressBarWithButtons] = {}
        self.ongoing_downloads_count = 0
        self.download_slot_available = Event()
        self.download_slot_available.set()
        self.prev_bar = None
        self.mutex = QMutex()
        self.cancelled = False

    def pause_or_resume(self):
        if not self.cancelled:
            for bar in self.progress_bars.values():
                bar.pause_button.click()
            self.anime_progress_bar.pause_or_resume()
        ProgressFunction.pause_or_resume(self)

    def cancel(self):
        if self.resume.is_set() and not self.cancelled:
            for bar in self.progress_bars.values():
                bar.cancel_button.click()
            self.downloaded_episode_count.cancelled = True
            self.anime_progress_bar.cancel()
            ProgressFunction.cancel(self)

    def update_anime_progress_bar(self, added: int):
        if self.anime_details.is_hls_download:
            self.update_anime_progress_bar_signal.emit(added)
        elif self.anime_details.skip_calculating_size:
            self.update_anime_progress_bar_signal.emit(added)
        else:
            added_rounded = round(added / IBYTES_TO_MBS_DIVISOR)
            self.update_anime_progress_bar_signal.emit(added_rounded)

    def clean_up_finished_download(self, episode_title: str):
        self.progress_bars.pop(episode_title)
        self.ongoing_downloads_count -= 1
        if self.ongoing_downloads_count < SETTINGS.max_simultaneous_downloads:
            self.download_slot_available.set()

    def update_eps_count_and_size(self, is_cancelled: bool, eps_file_path: str):
        hls_est_size = self.download_window.hls_est_size
        if is_cancelled:
            if not self.downloaded_episode_count.cancelled:
                self.downloaded_episode_count.total -= 1
            self.downloaded_episode_count.update_count(0)
            if hls_est_size:
                hls_est_size.update_count(0)
        else:
            self.downloaded_episode_count.update_count(1)
            if hls_est_size:
                eps_size = round(os.path.getsize(eps_file_path) / IBYTES_TO_MBS_DIVISOR)
                hls_est_size.update_count(eps_size)

    # Gogo's direct download link sometimes doesn't work, it returns a 301 - 308 status code meaning the resource has been moved, this attempts to redirect to that link
    # It is applied to Pahe too just in case and to make everything streamlined

    def run(self):
        ddls_or_segs_urls = self.anime_details.ddls_or_segs_urls
        for idx, ddl_or_seg_urls in enumerate(ddls_or_segs_urls):
            self.download_slot_available.wait()
            episode_title = self.anime_details.episode_title(idx)
            if self.anime_details.is_hls_download:
                episode_size_or_segs = len(ddl_or_seg_urls)
            else:
                (episode_size_or_segs, ddl_or_seg_urls) = Download.get_resource_length(
                    cast(str, ddl_or_seg_urls)
                )
                if episode_size_or_segs == 0:
                    continue
            # This is specifcally at this point instead of at the top cause of the above http request made in self.get_exact_episode_size such that if a user pauses or cancels as the request is in progress the input will be captured
            self.resume.wait()
            if self.cancelled:
                break
            self.mutex.lock()
            self.send_progress_bar_details.emit(
                episode_title,
                episode_size_or_segs,
                self.progress_bars,
                self.anime_details.is_hls_download,
            )
            self.mutex.unlock()
            while episode_title not in self.progress_bars:
                continue
            episode_progress_bar = self.progress_bars[episode_title]
            DownloadThread(
                self,
                ddl_or_seg_urls,
                episode_title,
                episode_size_or_segs,
                self.anime_details.site,
                self.anime_details.is_hls_download,
                self.anime_details.skip_calculating_size,
                self.anime_details.quality,
                self.anime_details.anime_folder_path,
                episode_progress_bar,
                self.clean_up_finished_download,
                self.anime_progress_bar,
                self.update_anime_progress_bar,
                self.update_eps_count_and_size,
                self.mutex,
            ).start()
            self.ongoing_downloads_count += 1
            if self.ongoing_downloads_count >= SETTINGS.max_simultaneous_downloads:
                self.download_slot_available.clear()


class DownloadThread(QThread):
    update_bars = pyqtSignal(int)
    finished = pyqtSignal(str)
    update_eps_count_and_hls_sizes = pyqtSignal(bool, str)
    update_bar_if_skipped_calculating_total_size = pyqtSignal(int)

    def __init__(
        self,
        parent: DownloadManagerThread,
        ddl_or_seg_urls: str | list[str],
        title: str,
        size: int,
        site: str,
        is_hls_download: bool,
        skipped_calculating_total_download_size: bool,
        hls_quality: str,
        download_folder: str,
        progress_bar: ProgressBarWithButtons,
        finished_callback: Callable,
        anime_progress_bar: ProgressBarWithoutButtons,
        update_anime_progress_bar: Callable,
        update_eps_count_and_hls_sizes: Callable,
        mutex: QMutex,
    ) -> None:
        super().__init__(parent)
        self.ddl_or_seg_urls = ddl_or_seg_urls
        self.title = title
        self.size = size
        self.download_folder = download_folder
        self.skipped_calculating_total_download_size = (
            skipped_calculating_total_download_size
        )
        self.site = site
        self.hls_quality = hls_quality
        self.is_hls_download = is_hls_download
        self.progress_bar = progress_bar
        self.anime_progress_bar = anime_progress_bar
        self.update_bars.connect(self.progress_bar.update_bar)
        if skipped_calculating_total_download_size:
            self.update_bar_if_skipped_calculating_total_size.connect(
                update_anime_progress_bar
            )
        else:
            self.update_bars.connect(update_anime_progress_bar)
        self.finished.connect(finished_callback)
        self.update_eps_count_and_hls_sizes.connect(update_eps_count_and_hls_sizes)
        self.mutex = mutex
        self.download: Download
        self.is_cancelled = False

    def cancel(self):
        self.download.cancel()
        divisor = 1 if self.is_hls_download else IBYTES_TO_MBS_DIVISOR
        new_maximum = self.anime_progress_bar.bar.maximum() - round(self.size / divisor)
        if new_maximum > 0:
            self.anime_progress_bar.bar.setMaximum(new_maximum)
        new_value = round(
            self.anime_progress_bar.bar.value()
            - round(self.progress_bar.bar.value() / divisor)
        )
        if new_value < 0:
            new_value = 0
        self.anime_progress_bar.bar.setValue(new_value)
        self.is_cancelled = True

    def run(self):
        if self.is_hls_download:
            self.ddl_or_seg_urls = cast(list[str], self.ddl_or_seg_urls)
            self.download = Download(
                self.ddl_or_seg_urls,
                self.title,
                self.download_folder,
                lambda x: self.update_bars.emit(x),
                is_hls_download=True,
            )
        else:
            self.ddl_or_seg_urls = cast(str, self.ddl_or_seg_urls)
            self.download = Download(
                self.ddl_or_seg_urls,
                self.title,
                self.download_folder,
                lambda x: self.update_bars.emit(x),
            )
        self.progress_bar.pause_callback = self.download.pause_or_resume
        self.progress_bar.cancel_callback = self.cancel

        self.download.start_download()
        self.mutex.lock()
        self.finished.emit(self.title)
        if self.skipped_calculating_total_download_size and not self.is_cancelled:
            self.update_bar_if_skipped_calculating_total_size.emit(1)
        self.update_eps_count_and_hls_sizes.emit(
            self.is_cancelled, self.download.file_path
        )
        self.mutex.unlock()


class GogoGetDownloadPageLinksThread(QThread):
    finished = pyqtSignal(AnimeDetails, list, list)
    hls_finished = pyqtSignal(AnimeDetails, list)

    def __init__(
        self,
        download_window: DownloadWindow,
        anime_details: AnimeDetails,
        callback: Callable[[AnimeDetails, list[str], list[list[str]]], None]
        | Callable[[AnimeDetails, list[str]], None],
    ):
        super().__init__(download_window)
        self.anime_details = anime_details
        self.hls_finished.connect(callback)
        self.finished.connect(callback)

    def run(self):
        page_content = gogo.get_anime_page_content(self.anime_details.anime.page_link)
        anime_id = gogo.extract_anime_id(page_content)
        download_page_links = gogo.get_download_page_links(
            self.anime_details.lacked_episode_numbers[0],
            self.anime_details.lacked_episode_numbers[-1],
            anime_id,
        )
        download_page_links = lacked_episodes(
            self.anime_details.lacked_episode_numbers, download_page_links
        )
        if self.anime_details.is_hls_download:
            self.hls_finished.emit(self.anime_details, download_page_links)
        else:
            self.finished.emit(self.anime_details, download_page_links, [])


class PaheGetEpisodePageInfo(QThread):
    finished = pyqtSignal(AnimeDetails, int, int, int, dict)

    def __init__(
        self,
        download_window: DownloadWindow,
        start_episode: int,
        end_episode: int,
        anime_details: AnimeDetails,
        finished_callback: Callable[
            [AnimeDetails, int, int, int, dict[str, Any]], None
        ],
    ):
        super().__init__(download_window)
        self.start_episode = start_episode
        self.end_episode = end_episode
        self.anime_details = anime_details
        self.finished.connect(finished_callback)

    def run(self):
        self.finished.emit(
            self.anime_details,
            *pahe.get_episode_pages_info(
                self.anime_details.anime.page_link, self.start_episode, self.end_episode
            ),
        )


class PaheGetEpisodePageLinksThread(QThread):
    finished = pyqtSignal(AnimeDetails, list)
    update_bar = pyqtSignal(int)

    def __init__(
        self,
        parent,
        anime_details: AnimeDetails,
        start_episode: int,
        end_episode: int,
        start_page_num: int,
        end_page_num: int,
        first_page: dict[str, Any],
        finished_callback: Callable[[AnimeDetails, list[str]], None],
        progress_bar: ProgressBarWithButtons,
    ):
        super().__init__(parent)
        self.anime_details = anime_details
        self.finished.connect(finished_callback)
        self.start_episode = start_episode
        self.end_episode = end_episode
        self.start_page_num = start_page_num
        self.end_page_num = end_page_num
        self.first_page = first_page
        self.progress_bar = progress_bar
        self.update_bar.connect(progress_bar.update_bar)

    def run(self):
        obj = pahe.GetEpisodePageLinks()
        self.progress_bar.pause_callback = obj.pause_or_resume
        self.progress_bar.cancel_callback = obj.cancel
        episode_page_links = obj.get_episode_page_links(
            self.start_episode,
            self.end_episode,
            self.start_page_num,
            self.end_page_num,
            self.first_page,
            self.anime_details.anime.page_link,
            cast(str, self.anime_details.anime.id),
            lambda x: self.update_bar.emit(x),
        )
        if not obj.cancelled:
            self.finished.emit(self.anime_details, episode_page_links)


class GetHlsLinksThread(QThread):
    finished = pyqtSignal(AnimeDetails, list)
    update_bar = pyqtSignal(int)

    def __init__(
        self,
        parent,
        episode_page_links: list[str],
        anime_details: AnimeDetails,
        progress_bar: ProgressBarWithButtons,
        finished_callback: Callable[[AnimeDetails, list[str]], None],
    ):
        super().__init__(parent)
        self.anime_details = anime_details
        self.episode_page_links = episode_page_links
        self.progress_bar = progress_bar
        self.finished.connect(finished_callback)
        self.update_bar.connect(self.progress_bar.update_bar)

    def run(self):
        obj = gogo.GetHlsLinks()
        self.progress_bar.pause_callback = obj.pause_or_resume
        self.progress_bar.cancel_callback = obj.cancel
        hls_links = obj.get_hls_links(self.episode_page_links, self.update_bar.emit)
        if not obj.cancelled:
            self.finished.emit(self.anime_details, hls_links)


class GogoHlsGetMatchedQualityLinksThread(QThread):
    finished = pyqtSignal(AnimeDetails, list)
    update_bar = pyqtSignal(int)

    def __init__(
        self,
        parent,
        hls_links: list[str],
        anime_details: AnimeDetails,
        progress_bar: ProgressBarWithButtons,
        finished_callback: Callable[[AnimeDetails, list[str]], None],
    ):
        super().__init__(parent)
        self.hls_links = hls_links
        self.anime_details = anime_details
        self.progress_bar = progress_bar
        self.finished.connect(finished_callback)
        self.update_bar.connect(self.progress_bar.update_bar)

    def run(self):
        obj = gogo.GetHlsMatchedQualityLinks()
        self.progress_bar.pause_callback = obj.pause_or_resume
        self.progress_bar.cancel_callback = obj.cancel
        matched_links = obj.get_hls_matched_quality_links(
            self.hls_links, self.anime_details.quality, self.update_bar.emit
        )
        if not obj.cancelled:
            self.finished.emit(self.anime_details, matched_links)


class GogoHlsGetSegmentsUrlsThread(QThread):
    finished = pyqtSignal(AnimeDetails)
    update_bar = pyqtSignal(int)

    def __init__(
        self,
        parent,
        matched_links: list[str],
        anime_details: AnimeDetails,
        progress_bar: ProgressBarWithButtons,
        finished_callback: Callable[[AnimeDetails], None],
    ):
        super().__init__(parent)
        self.matched_links = matched_links
        self.anime_details = anime_details
        self.progress_bar = progress_bar
        self.finished.connect(finished_callback)
        self.update_bar.connect(self.progress_bar.update_bar)

    def run(self):
        obj = gogo.GetHlsSegmentsUrls()
        self.progress_bar.pause_callback = obj.pause_or_resume
        self.progress_bar.cancel_callback = obj.cancel
        self.anime_details.ddls_or_segs_urls = obj.get_hls_segments_urls(
            self.matched_links, self.update_bar.emit
        )
        if not obj.cancelled:
            self.finished.emit(self.anime_details)


class PaheGetDownloadPageThread(QThread):
    finished = pyqtSignal(AnimeDetails, list, list)
    update_bar = pyqtSignal(int)

    def __init__(
        self,
        parent,
        anime_details: AnimeDetails,
        episode_page_links: list[str],
        finished_callback: Callable[[AnimeDetails, list, list], None],
        progress_bar: ProgressBarWithButtons,
    ):
        super().__init__(parent)
        self.anime_details = anime_details
        self.episode_page_links = episode_page_links
        self.finished.connect(finished_callback)
        self.progress_bar = progress_bar
        self.update_bar.connect(progress_bar.update_bar)

    def run(self):
        obj = pahe.GetPahewinPageLinks()
        self.progress_bar.pause_callback = obj.pause_or_resume
        self.progress_bar.cancel_callback = obj.cancel
        d_page, d_info = obj.get_pahewin_page_links_and_info(
            self.episode_page_links, self.update_bar.emit
        )
        if not obj.cancelled:
            return self.finished.emit(self.anime_details, d_page, d_info)


class GetDirectDownloadLinksThread(QThread):
    finished = pyqtSignal(AnimeDetails)
    update_bar = pyqtSignal(int)

    def __init__(
        self,
        download_window: DownloadWindow,
        download_page_links: list[str] | list[list[str]],
        download_info: list[list[str]],
        anime_details: AnimeDetails,
        finished_callback: Callable[[AnimeDetails], None],
        progress_bar: ProgressBarWithButtons,
    ):
        super().__init__(download_window)
        self.download_window = download_window
        self.download_page_links = download_page_links
        self.download_info = download_info
        self.anime_details = anime_details
        self.finished.connect(finished_callback)
        self.progress_bar = progress_bar
        self.update_bar.connect(progress_bar.update_bar)

    def run(self):
        if self.anime_details.site == PAHE:
            bound_links, bound_info = pahe.bind_sub_or_dub_to_link_info(
                self.anime_details.sub_or_dub,
                cast(list[list[str]], self.download_page_links),
                self.download_info,
            )
            bound_links, bound_info = pahe.bind_quality_to_link_info(
                self.anime_details.quality, bound_links, bound_info
            )
            self.anime_details.download_info = bound_info
            obj = pahe.GetDirectDownloadLinks()
            self.progress_bar.pause_callback = obj.pause_or_resume
            self.progress_bar.cancel_callback = obj.cancel
            self.anime_details.ddls_or_segs_urls = obj.get_direct_download_links(
                bound_links, lambda x: self.update_bar.emit(x)
            )
        else:
            obj = gogo.GetDirectDownloadLinks()
            self.progress_bar.pause_callback = obj.pause_or_resume
            self.progress_bar.cancel_callback = obj.cancel
            self.anime_details.ddls_or_segs_urls = obj.get_direct_download_links(
                cast(list[str], self.download_page_links),
                self.anime_details.quality,
                lambda x: self.update_bar.emit(x),
            )

        if len(self.anime_details.ddls_or_segs_urls) < len(self.download_page_links):
            self.download_window.main_window.tray_icon.make_notification(
                "Error",
                f"Failed to find some {'hls' if self.anime_details.is_hls_download else 'direct download'} links for {self.anime_details.anime.title}",
                False,
                None,
            )
        if not self.anime_details.ddls_or_segs_urls:
            return
        if not obj.cancelled:
            self.finished.emit(self.anime_details)


class GogoCalculateDownloadSizes(QThread):
    finished = pyqtSignal(AnimeDetails)
    update_bar = pyqtSignal(int)

    def __init__(
        self,
        parent: QWidget,
        anime_details: AnimeDetails,
        finished_callback: Callable[[AnimeDetails], None],
        progress_bar: ProgressBarWithButtons,
    ):
        super().__init__(parent)
        self.anime_details = anime_details
        self.finished.connect(finished_callback)
        self.progress_bar = progress_bar
        self.update_bar.connect(progress_bar.update_bar)

    def run(self):
        obj = gogo.CalculateTotalDowloadSize()
        self.progress_bar.pause_callback = obj.pause_or_resume
        self.progress_bar.cancel_callback = obj.cancel
        self.anime_details.total_download_size = obj.calculate_total_download_size(
            cast(list[str], self.anime_details.ddls_or_segs_urls),
            lambda x: self.update_bar.emit(x),
            True,
        )
        if not obj.cancelled:
            self.finished.emit(self.anime_details)


class AutoDownloadThread(QThread):
    initate_download_pipeline = pyqtSignal(AnimeDetails)
    clean_out_auto_download_thread_signal = pyqtSignal()

    def __init__(
        self,
        download_window: DownloadWindow,
        titles: list[str],
        tray_icon: QSystemTrayIcon,
        clean_out_auto_download_thread_slot: Callable[[], None],
    ):
        super().__init__(download_window)
        self.anime_titles = titles
        self.download_window = download_window
        self.initate_download_pipeline.connect(
            self.download_window.initiate_download_pipeline
        )
        self.tray_icon = tray_icon
        self.clean_out_auto_download_thread_signal.connect(
            clean_out_auto_download_thread_slot
        )

    def run(self):
        queued: list[str] = []
        for title in self.anime_titles:
            anime: Anime
            site = SETTINGS.auto_download_site
            if site == PAHE:
                result = self.pahe_fetch_anime_obj(title)
                if not result:
                    result = self.gogo_fetch_anime_obj(title)
                    if not result:
                        continue
                    site = GOGO
            else:
                result = self.gogo_fetch_anime_obj(title)
                if not result:
                    result = self.pahe_fetch_anime_obj(title)
                    if not result:
                        continue
                    site = PAHE
            anime = result
            anime_details = AnimeDetails(anime, site)
            start_eps = anime_details.haved_end if anime_details.haved_end else 1
            anime_details.lacked_episode_numbers = lacked_episode_numbers(
                start_eps, anime_details.episode_count, anime_details.haved_episodes
            )
            if not anime_details.lacked_episode_numbers:
                haved_end = anime_details.haved_end
                if anime_details.metadata.airing_status == "FINISHED" and (
                    haved_end and haved_end >= anime_details.episode_count
                ):
                    self.download_window.main_window.settings_window.tracked_anime.remove_anime(
                        anime_details.anime.title
                    )
                    self.download_window.main_window.tray_icon.make_notification(
                        "Finished Tracking",
                        f"You have the final episode of {title} and it has finished airing so I have removed it from your tracking list",
                        True,
                    )
                continue
            if anime_details.sub_or_dub == DUB and not anime_details.dub_available:
                self.download_window.main_window.tray_icon.make_notification(
                    "Error",
                    f"Failed to find dub for {anime_details.anime.title}",
                    False,
                    None,
                )
                continue
            queued.append(anime_details.anime.title)
            self.initate_download_pipeline.emit(anime_details)
        if queued:
            all_str = ", ".join(queued)
            self.download_window.main_window.tray_icon.make_notification(
                "Queued new episodes",
                all_str,
                False,
                self.download_window.main_window.switch_to_download_window,
            )
        self.clean_out_auto_download_thread_signal.emit()

    def pahe_fetch_anime_obj(self, title: str) -> Anime | None:
        results = pahe.search(title)
        for result in results:
            res_title, page_link, anime_id = pahe.extract_anime_title_page_link_and_id(
                result
            )
            if sanitise_title(res_title.lower(), True) == sanitise_title(
                title.lower(), True
            ):
                return Anime(title, page_link, anime_id)
        return None

    def gogo_fetch_anime_obj(self, title: str) -> Anime | None:
        results = gogo.search(title)
        for res_title, page_link in results:
            if sanitise_title(res_title.lower(), True) == sanitise_title(
                title.lower(), True
            ):
                return Anime(title, page_link, None)
        return None



from typing import Callable, cast

from PyQt6.QtCore import QThread, Qt
from PyQt6.QtGui import QAction, QCloseEvent, QGuiApplication, QIcon, QScreen
from PyQt6.QtWidgets import (
    QApplication,
    QMainWindow,
    QMenu,
    QStackedWidget,
    QSystemTrayIcon,
    QWidget,
    QMessageBox,
)
from senpwai.utils.classes import SETTINGS, Anime, AnimeDetails
from senpwai.utils.static import (
    MINIMISED_TO_TRAY_ARG,
    SENPWAI_ICON_PATH,
    TASK_COMPLETE_ICON_PATH,
    UPDATE_ICON_PATH,
)

from senpwai.windows.about import AboutWindow
from senpwai.windows.abstracts import AbstractWindow
from senpwai.windows.chosen_anime import ChosenAnimeWindow, MakeAnimeDetailsThread
from senpwai.windows.download import DownloadWindow
from senpwai.windows.misc import (
    CheckIfUpdateAvailableThread,
    NoFFmpegWindow,
    UpdateWindow,
)
from senpwai.windows.search import SearchWindow
from senpwai.windows.settings import SettingsWindow
from senpwai.windows.abstracts import NavBarButton


class MainWindow(QMainWindow):
    def __init__(self, app: QApplication):
        super().__init__()
        self.set_bckg_img = lambda img_path: self.setStyleSheet(
            f"QMainWindow{{border-image: url({img_path}) 0 0 0 0 stretch stretch;}}"
        )
        self.app = app
        self.center_window()
        self.senpwai_icon = QIcon(SENPWAI_ICON_PATH)
        self.search_window = SearchWindow(self)
        self.download_window = DownloadWindow(self)
        self.tray_icon = TrayIcon(self)
        self.tray_icon.show()
        self.settings_window = SettingsWindow(self)
        self.about_window = AboutWindow(self)
        CheckIfUpdateAvailableThread(self, self.handle_update_check_result).start()
        self.stacked_windows = QStackedWidget(self)
        # The widget that is added first is implicitly set as the current widget
        self.stacked_windows.addWidget(self.search_window)
        self.set_bckg_img(self.search_window.bckg_img_path)
        self.stacked_windows.addWidget(self.download_window)
        self.stacked_windows.addWidget(self.settings_window)
        self.stacked_windows.addWidget(self.about_window)
        self.setCentralWidget(self.stacked_windows)
        self.setup_chosen_anime_window_thread: QThread | None = None
        self.download_window.start_auto_download()

    def closeEvent(self, a0: QCloseEvent | None) -> None:
        if not a0:
            return
        if self.download_window.is_downloading():
            message_box = QMessageBox(self)
            message_box.setIcon(QMessageBox.Icon.Warning)
            message_box.setStyleSheet("color: black")
            message_box.setText(
                "You have ongoing downloads, are you sure you want to exit?"
            )
            message_box.setStandardButtons(
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
            )
            message_box.setDefaultButton(QMessageBox.StandardButton.No)
            message_box.exec()
            if message_box.result() == QMessageBox.StandardButton.No:
                return a0.ignore()
        a0.accept()

    def show_with_settings(self, args: list[str]):
        if MINIMISED_TO_TRAY_ARG in args:
            if SETTINGS.start_in_fullscreen:
                self.setWindowState(self.windowState() | Qt.WindowState.WindowMaximized)
            return self.hide()
        elif SETTINGS.start_in_fullscreen:
            self.showMaximized()
        else:
            self.showNormal()
            self.center_window()
            self.setWindowState(Qt.WindowState.WindowActive)

    def show(self):
        self.activateWindow()
        if Qt.WindowState.WindowMaximized in self.windowState():
            return self.showMaximized()
        self.showNormal()
        self.center_window()

    def handle_update_check_result(self, result: tuple[bool, str, str, str]):
        is_available, download_url, file_name, update_info = result
        if not is_available:
            return
        self.update_window = UpdateWindow(self, download_url, file_name, update_info)
        self.stacked_windows.addWidget(self.update_window)
        update_icon = NavBarButton(UPDATE_ICON_PATH, self.switch_to_update_window)
        self.search_window.nav_bar_layout.addWidget(update_icon)

    def switch_to_update_window(self):
        self.switch_to_window(self.update_window)

    def center_window(self) -> None:
        screen_geometry = cast(
            QScreen, QGuiApplication.primaryScreen()
        ).availableGeometry()
        x = (screen_geometry.width() - self.width()) // 2
        self.move(x, 0)

    def setup_and_switch_to_chosen_anime_window(self, anime: Anime, site: str):
        # This if statement prevents error: "QThread: Destroyed while thread is still running" that happens when more than one thread is spawned
        # When a user clicks more than one ResultButton quickly causing the reference to the original thread to be overwridden hence garbage collected/destroyed
        if self.setup_chosen_anime_window_thread is None:
            self.search_window.loading.start()
            self.search_window.bottom_section_stacked_widgets.setCurrentWidget(
                self.search_window.loading
            )
            self.setup_chosen_anime_window_thread = MakeAnimeDetailsThread(
                self, anime, site
            )
            self.setup_chosen_anime_window_thread.finished.connect(
                lambda anime_details: self.make_chosen_anime_window(anime_details)
            )
            self.setup_chosen_anime_window_thread.start()

    def make_chosen_anime_window(self, anime_details: AnimeDetails):
        self.setup_chosen_anime_window_thread = None
        self.search_window.bottom_section_stacked_widgets.setCurrentWidget(
            self.search_window.results_widget
        )
        self.search_window.loading.stop()
        chosen_anime_window = ChosenAnimeWindow(self, anime_details)
        self.stacked_windows.addWidget(chosen_anime_window)
        self.set_bckg_img(chosen_anime_window.bckg_img_path)
        self.stacked_windows.setCurrentWidget(chosen_anime_window)
        self.setup_chosen_anime_window_thread = None

    def switch_to_pahe(self, anime_title: str, initiator: QWidget):
        self.search_window.search_bar.setText(anime_title)
        self.search_window.pahe_search_button.animateClick()
        self.switch_to_search_window()
        self.stacked_windows.removeWidget(initiator)
        initiator.deleteLater()

    def switch_to_gogo(self, anime_title: str, initiator: QWidget):
        self.search_window.search_bar.setText(anime_title)
        self.search_window.gogo_search_button.animateClick()
        self.switch_to_search_window()
        self.stacked_windows.removeWidget(initiator)
        initiator.deleteLater()

    def switch_to_window(self, target_window: AbstractWindow):
        if self.stacked_windows.currentWidget() != target_window:
            self.set_bckg_img(target_window.bckg_img_path)
            self.stacked_windows.setCurrentWidget(target_window)

    def switch_to_search_window(self):
        self.switch_to_window(self.search_window)
        self.search_window.search_bar.setFocus()
        self.search_window.on_focus()

    def switch_to_download_window(self):
        self.switch_to_window(self.download_window)

    def switch_to_settings_window(self):
        self.switch_to_window(self.settings_window)

    def switch_to_about_window(self):
        self.switch_to_window(self.about_window)

    def create_and_switch_to_no_ffmpeg_window(self, anime_details: AnimeDetails):
        no_ffmpeg_window = NoFFmpegWindow(self, anime_details)
        self.stacked_windows.addWidget(no_ffmpeg_window)
        self.switch_to_window(no_ffmpeg_window)


class TrayIcon(QSystemTrayIcon):
    def __init__(self, main_window: MainWindow):
        super().__init__(main_window.senpwai_icon, main_window)
        self.task_complete_icon = QIcon(TASK_COMPLETE_ICON_PATH)
        self.main_window = main_window
        self.setToolTip("Senpwai")
        self.activated.connect(self.on_tray_icon_click)
        self.context_menu = QMenu("Senpwai", main_window)
        check_for_new_episodes_action = QAction(
            "Check for new episodes", self.context_menu
        )
        check_for_new_episodes_action.triggered.connect(
            main_window.download_window.start_auto_download
        )
        search_action = QAction("Search", self.context_menu)
        search_action.triggered.connect(main_window.switch_to_search_window)
        search_action.triggered.connect(main_window.show)
        downloads_action = QAction("Downloads", self.context_menu)
        downloads_action.triggered.connect(main_window.switch_to_download_window)
        downloads_action.triggered.connect(main_window.show)
        exit_action = QAction("Exit", self.context_menu)
        exit_action.triggered.connect(main_window.app.quit)
        self.context_menu.addAction(check_for_new_episodes_action)
        self.context_menu.addAction(search_action)
        self.context_menu.addAction(downloads_action)
        self.context_menu.addAction(exit_action)
        self.messageClicked.connect(main_window.show)
        self.setContextMenu(self.context_menu)

    def focus_or_hide_window(self):
        if not self.main_window.isVisible():
            return self.main_window.show()
        self.main_window.hide()

    def on_tray_icon_click(self, reason: QSystemTrayIcon.ActivationReason):
        if reason == QSystemTrayIcon.ActivationReason.Context:
            self.context_menu.show()
        elif reason == QSystemTrayIcon.ActivationReason.Trigger:
            self.focus_or_hide_window()

    def make_notification(
        self,
        title: str,
        msg: str,
        sth_completed: bool,
        onclick: Callable[[], None] | None = None,
    ):
        if SETTINGS.allow_notifications:
            if onclick:
                self.messageClicked.connect(onclick)
            if sth_completed:
                return self.showMessage(title, msg, self.task_complete_icon, 5000)
            self.showMessage(title, msg, msecs=5000)



from typing import TYPE_CHECKING, Callable

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QHBoxLayout,
    QVBoxLayout,
    QWidget,
)
from senpwai.utils.static import (
    ABOUT_ICON_PATH,
    DOWNLOADS_ICON_PATH,
    SEARCH_ICON_PATH,
    SETTINGS_ICON_PATH,
)
from senpwai.utils.widgets import IconButton, Icon

# https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports/3957388#39757388
if TYPE_CHECKING:
    from senpwai.windows.main import MainWindow


class NavBarButton(IconButton):
    def __init__(self, icon_path: str, switch_to_window_callable: Callable):
        super().__init__(Icon(50, 50, icon_path), 1.15)
        self.clicked.connect(switch_to_window_callable)
        self.setStyleSheet(
            """
            QPushButton {
                    border-radius: 21px;
                    background-color: black;
                }"""
        )


class AbstractWindow(QWidget):
    def __init__(self, main_window: 'MainWindow', bckg_img_path: str):
        super().__init__(main_window)
        self.bckg_img_path = bckg_img_path
        self.full_layout = QHBoxLayout()
        nav_bar_widget = QWidget()
        self.nav_bar_layout = QVBoxLayout()

        self.search_window_button = NavBarButton(
            SEARCH_ICON_PATH, main_window.switch_to_search_window
        )
        self.download_window_button = NavBarButton(
            DOWNLOADS_ICON_PATH, main_window.switch_to_download_window
        )
        self.settings_window_button = NavBarButton(
            SETTINGS_ICON_PATH, main_window.switch_to_settings_window
        )
        self.about_window_button = NavBarButton(
            ABOUT_ICON_PATH, main_window.switch_to_about_window
        )
        self.nav_bar_buttons = [
            self.search_window_button,
            self.download_window_button,
            self.settings_window_button,
            self.about_window_button,
        ]
        for button in self.nav_bar_buttons:
            self.nav_bar_layout.addWidget(button)
        self.nav_bar_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        nav_bar_widget.setLayout(self.nav_bar_layout)
        self.full_layout.addWidget(nav_bar_widget)


class AbstractTemporaryWindow(AbstractWindow):
    def __init__(self, main_window: 'MainWindow', bckg_img_path: str):
        super().__init__(main_window, bckg_img_path)
        for button in self.nav_bar_buttons:
            button.clicked.connect(
                lambda: main_window.stacked_windows.removeWidget(self)
            )
            button.clicked.connect(self.deleteLater)



from random import choice as randomchoice
from time import sleep as timesleep
from typing import TYPE_CHECKING, cast

from PyQt6.QtCore import QEvent, QObject, Qt, QThread, QTimer, pyqtSignal
from PyQt6.QtGui import QKeyEvent
from PyQt6.QtWidgets import (
    QHBoxLayout,
    QLayoutItem,
    QLineEdit,
    QScrollBar,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)
from senpwai.scrapers import gogo, pahe
from senpwai.utils.classes import Anime
from senpwai.utils.scraper import CLIENT
from senpwai.utils.static import (
    ANILIST_API_ENTRYPOINY,
    BUNSHIN_POOF_AUDIO_PATH,
    GIGACHAD_AUDIO_PATH,
    GOGO,
    GOGO_HOVER_COLOR,
    GOGO_NORMAL_COLOR,
    GOGO_PRESSED_COLOR,
    IS_CHRISTMAS,
    KAGE_BUNSHIN_AUDIO_PATH,
    L_ANIME,
    LOADING_ANIMATION_PATH,
    MERRY_CHRISMASU_AUDIO_PATH,
    PAHE,
    PAHE_HOVER_COLOR,
    PAHE_NORMAL_COLOR,
    PAHE_PRESSED_COLOR,
    RANDOM_MACOT_ICON_PATH,
    SADGE_PIECE_PATH,
    SEARCH_WINDOW_BCKG_IMAGE_PATH,
    SEN_ANILIST_ID,
    SEN_FAVOURITE_AUDIO_PATH,
    TOKI_WA_UGOKI_DASU_AUDIO_PATH,
    W_ANIME,
    WHAT_DA_HELL_AUDIO_PATH,
    ZA_WARUDO_AUDIO_PATH,
)
from senpwai.utils.widgets import (
    AnimationAndText,
    AudioPlayer,
    Icon,
    IconButton,
    OutlinedButton,
    ScrollableSection,
    StyledButton,
    set_minimum_size_policy,
)

from senpwai.windows.abstracts import AbstractWindow
# https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports/3957388#39757388
if TYPE_CHECKING:
    from senpwai.windows.main import MainWindow

class SearchWindow(AbstractWindow):
    def __init__(self, main_window: 'MainWindow'):
        super().__init__(main_window, SEARCH_WINDOW_BCKG_IMAGE_PATH)
        self.main_window = main_window
        main_widget = QWidget()
        main_layout = QVBoxLayout()

        mascot_button = IconButton(Icon(117, 100, RANDOM_MACOT_ICON_PATH), 1)
        mascot_button.setToolTip("Goofy 🗿")

        mascot_button.clicked.connect(
            AudioPlayer(self, SEN_FAVOURITE_AUDIO_PATH, volume=60).play
        )
        mascot_button.clicked.connect(FetchFavouriteThread(self).start)
        self.search_bar = SearchBar(self)
        self.get_search_bar_text = lambda: self.search_bar.text()
        self.search_bar.setMinimumHeight(60)

        search_bar_and_mascot_widget = QWidget()
        search_bar_and_mascot_layout = QVBoxLayout()
        search_bar_and_mascot_layout.addWidget(
            mascot_button, alignment=Qt.AlignmentFlag.AlignHCenter
        )
        search_bar_and_mascot_layout.addWidget(self.search_bar)
        search_bar_and_mascot_layout.setSpacing(0)
        search_bar_and_mascot_widget.setLayout(search_bar_and_mascot_layout)
        main_layout.addWidget(search_bar_and_mascot_widget)

        search_buttons_widget = QWidget()
        search_buttons_layout = QHBoxLayout()
        self.pahe_search_button = SearchButton(self, PAHE)
        set_minimum_size_policy(self.pahe_search_button)
        self.gogo_search_button = SearchButton(self, GOGO)
        set_minimum_size_policy(self.gogo_search_button)
        search_buttons_layout.addWidget(self.pahe_search_button)
        search_buttons_layout.addWidget(self.gogo_search_button)
        search_buttons_widget.setLayout(search_buttons_layout)
        main_layout.addWidget(search_buttons_widget)
        self.bottom_section_stacked_widgets = QStackedWidget()

        self.results_layout = QVBoxLayout()
        self.results_widget = ScrollableSection(self.results_layout)
        self.res_wid_hor_scroll_bar = cast(
            QScrollBar, self.results_widget.horizontalScrollBar()
        )

        self.loading = AnimationAndText(
            LOADING_ANIMATION_PATH, 250, 300, "Loading.. .", 1, 48, 50
        )
        self.anime_not_found = AnimationAndText(
            SADGE_PIECE_PATH, 400, 300, ":( couldn't find that anime ", 1, 48, 50
        )
        self.bottom_section_stacked_widgets.addWidget(self.results_widget)
        self.bottom_section_stacked_widgets.addWidget(self.loading)
        self.bottom_section_stacked_widgets.addWidget(self.anime_not_found)
        self.bottom_section_stacked_widgets.setCurrentWidget(self.results_widget)
        main_layout.addWidget(self.bottom_section_stacked_widgets)
        self.search_thread = None
        main_widget.setLayout(main_layout)
        self.full_layout.addWidget(main_widget)
        self.setLayout(self.full_layout)
        # We use a timer instead of calling setFocus normally cause apparently Qt wont really set the widget in focus if the widget isn't shown on screen/rendered,
        # So we gotta wait a bit first till the UI is rendered. StackOverflow Comment link: https://stackoverflow.com/questions/52853701/set-focus-on-button-in-app-with-group-boxes#comment92652037_52858926
        QTimer.singleShot(0, self.search_bar.setFocus)

    # Qt pushes the horizontal scroll bar to the center automatically sometimes
    def fix_hor_scroll_bar(self):
        self.res_wid_hor_scroll_bar.setValue(self.res_wid_hor_scroll_bar.minimum())

    def on_focus(self):
        self.search_bar.setFocus()
        self.fix_hor_scroll_bar()

    def search_anime(self, anime_title: str, site: str) -> None:
        if not anime_title:
            return
        if self.search_thread:
            self.search_thread.quit()
        prev_top_was_anime_not_found = (
            self.bottom_section_stacked_widgets.currentWidget() == self.anime_not_found
        )
        self.loading.start()
        self.bottom_section_stacked_widgets.setCurrentWidget(self.loading)
        if prev_top_was_anime_not_found:
            self.anime_not_found.stop()
        for idx in reversed(range(self.results_layout.count())):
            item = cast(
                QWidget, cast(QLayoutItem, self.results_layout.itemAt(idx)).widget()
            ).deleteLater()
            self.results_layout.removeItem(item)
        self.search_thread = SearchThread(self, anime_title, site)
        anime_title_lower = anime_title.lower()
        is_naruto = "naruto" in anime_title_lower
        if "one piece" in anime_title_lower:
            AudioPlayer(self, SEN_FAVOURITE_AUDIO_PATH, volume=100).play()
        elif "jojo" in anime_title_lower:
            AudioPlayer(self, ZA_WARUDO_AUDIO_PATH, 100).play()
            for _ in range(180):
                self.main_window.app.processEvents()
                timesleep(0.01)
            for x in range(20):
                self.main_window.app.processEvents()
                timesleep(x * 0.01)
            timesleep(2)
            AudioPlayer(self, TOKI_WA_UGOKI_DASU_AUDIO_PATH, 100).play()
            timesleep(1.8)
        elif anime_title_lower in W_ANIME:
            AudioPlayer(self, GIGACHAD_AUDIO_PATH, 25).play()
        elif anime_title_lower in L_ANIME:
            AudioPlayer(self, WHAT_DA_HELL_AUDIO_PATH, 100).play()
        elif is_naruto:
            self.kage_bunshin_no_jutsu = AudioPlayer(
                self, KAGE_BUNSHIN_AUDIO_PATH, volume=50
            )
            self.kage_bunshin_no_jutsu.play()
        # Has to be the last check to avoid faiing to set self.kage_bunshin_no_jutsu incase is_naruto == True
        elif IS_CHRISTMAS:
            AudioPlayer(self, MERRY_CHRISMASU_AUDIO_PATH, 30).play()

        if is_naruto:
            self.search_thread.finished.connect(self.start_naruto_results_thread)
        else:
            self.search_thread.finished.connect(self.show_results)
        self.search_thread.start()

    def start_naruto_results_thread(self, site: str, results: list[Anime]):
        NarutoResultsThread(self, site, results).start()

    def play_bunshin_poof(self):
        AudioPlayer(self, BUNSHIN_POOF_AUDIO_PATH, 10).play()

    def show_results(self, site: str, results: list[Anime]):
        if not results:
            self.anime_not_found.start()
            self.bottom_section_stacked_widgets.setCurrentWidget(self.anime_not_found)
        else:
            self.bottom_section_stacked_widgets.setCurrentWidget(self.results_widget)
            for result in results:
                button = ResultButton(result, self.main_window, self, site, 9, 48)
                self.results_layout.addWidget(button)
        self.loading.stop()
        self.search_thread = None

    def make_naruto_result_button(self, result: Anime, site: str):
        button = ResultButton(result, self.main_window, self, site, 9, 48)
        self.results_layout.addWidget(button)


class NarutoResultsThread(QThread):
    send_result = pyqtSignal(Anime, str)
    stop_loading_animation = pyqtSignal()
    start_anime_not_found_animation = pyqtSignal()
    set_curr_wid = pyqtSignal(QWidget)
    play_bunshin = pyqtSignal()

    def __init__(self, search_window: SearchWindow, site: str, results: list[Anime]):
        super().__init__(search_window)
        self.search_window = search_window
        self.results = results
        self.site = site
        self.bunshin_poof = AudioPlayer(search_window, BUNSHIN_POOF_AUDIO_PATH)
        self.send_result.connect(search_window.make_naruto_result_button)
        self.stop_loading_animation.connect(search_window.loading.stop)
        self.start_anime_not_found_animation.connect(
            search_window.anime_not_found.start
        )
        self.set_curr_wid.connect(
            search_window.bottom_section_stacked_widgets.setCurrentWidget
        )
        self.play_bunshin.connect(search_window.play_bunshin_poof)

    def run(self):
        while self.search_window.kage_bunshin_no_jutsu.isPlaying():
            pass
        if not self.results:
            self.start_anime_not_found_animation.emit()
            self.set_curr_wid.emit(self.search_window.anime_not_found)
        else:
            self.stop_loading_animation.emit()
            self.set_curr_wid.emit(self.search_window.results_widget)
            for idx, result in enumerate(self.results):
                self.send_result.emit(result, self.site)
                if idx <= 5:
                    self.play_bunshin.emit()
                    timesleep(0.35)
            self.search_window.search_thread = None


class FetchFavouriteThread(QThread):
    def __init__(self, search_window: SearchWindow) -> None:
        super().__init__(search_window)
        self.search_window = search_window

    def run(self):
        favourite = self.get_random_sen_favourite()
        if not favourite:
            return
        self.type_write_favourite_in_search_bar(favourite)

    def type_write_favourite_in_search_bar(self, favourite_name: str):
        self.search_window.search_bar.clear()
        for idx, _ in enumerate(favourite_name):
            self.search_window.search_bar.setText(favourite_name[: idx + 1])
            timesleep(0.1)

    def get_random_sen_favourite(self) -> str | None:
        page = randomchoice((1, 2))
        query = """
        query getUserFavourite($id: Int, $page: Int){
        User(id: $id) {
            favourites{
            anime(page: $page){
                nodes{
                title{
                    romaji
                }
                }
                pageInfo {
                lastPage
                perPage
                total
                }
            }
            }
        }
        }
        """
        response = CLIENT.post(
            ANILIST_API_ENTRYPOINY,
            json={"query": query, "variables": {"id": SEN_ANILIST_ID, "page": page}},
            headers=CLIENT.append_headers({"Content-Type": "application/json"}),
        )
        if response.status_code != 200:
            return None
        data = response.json()
        favourite_anime = data["data"]["User"]["favourites"]["anime"]["nodes"]
        count = len(favourite_anime)
        if count <= 0:
            return None
        chosen = randomchoice(favourite_anime)
        anime_title = chosen["title"]["romaji"]
        return anime_title


class SearchBar(QLineEdit):
    def __init__(self, search_window: SearchWindow):
        super().__init__()
        self.search_window = search_window
        self.setPlaceholderText("Enter anime title")
        self.installEventFilter(self)
        self.setStyleSheet(
            """
            QLineEdit{
                border: 1px solid black;
                border-radius: 15px;
                padding: 5px;
                color: black;
                font-size: 30px;
                font-family: "Berlin Sans FB Demi";
            }
        """
        )

    def eventFilter(self, a0: QObject | None, a1: QEvent | None):
        if isinstance(a1, QKeyEvent):
            if a0 == self and a1.type() == a1.Type.KeyPress:
                if a1.key() == Qt.Key.Key_Enter or a1.key() == Qt.Key.Key_Return:
                    self.search_window.pahe_search_button.animateClick()
                elif a1.key() == Qt.Key.Key_Tab:
                    first_button = self.search_window.results_layout.itemAt(0)
                    if first_button:
                        cast(QWidget, first_button.widget()).setFocus()
                    else:
                        self.search_window.gogo_search_button.animateClick()
                    return True
        return super().eventFilter(a0, a1)


class SearchButton(StyledButton):
    def __init__(self, window: SearchWindow, site: str):
        if site == PAHE:
            super().__init__(
                window,
                40,
                "black",
                PAHE_NORMAL_COLOR,
                PAHE_HOVER_COLOR,
                PAHE_PRESSED_COLOR,
            )
            self.setText("Animepahe")
            self.setToolTip("Stable 🌲")
        else:
            super().__init__(
                window,
                40,
                "black",
                GOGO_NORMAL_COLOR,
                GOGO_HOVER_COLOR,
                GOGO_PRESSED_COLOR,
            )
            self.setText("Gogoanime")
            self.setToolTip("Experimental 🧪")
        self.clicked.connect(
            lambda: window.search_anime(window.get_search_bar_text(), site)
        )


class ResultButton(OutlinedButton):
    def __init__(
        self,
        anime: Anime,
        main_window: 'MainWindow',
        search_window: SearchWindow,
        site: str,
        paint_x: int,
        paint_y: int,
    ):
        self.search_window = search_window
        if site == PAHE:
            hover_color = PAHE_NORMAL_COLOR
            pressed_color = PAHE_HOVER_COLOR
        else:
            hover_color = GOGO_NORMAL_COLOR
            pressed_color = GOGO_HOVER_COLOR
        super().__init__(
            paint_x,
            paint_y,
            None,
            40,
            "white",
            "transparent",
            hover_color,
            pressed_color,
            21,
        )
        self.setText(anime.title)
        self.setStyleSheet(
            self.styleSheet()
            + """
                           QPushButton{
                           text-align: left;
                           border: none;
                           }"""
        )
        self.style_sheet_buffer = self.styleSheet()
        self.focused_sheet = (
            self.style_sheet_buffer
            + f"""
                    QPushButton{{
                        background-color: {hover_color};
        }}"""
        )
        self.clicked.connect(
            lambda: main_window.setup_and_switch_to_chosen_anime_window(anime, site)
        )
        self.installEventFilter(self)

    def eventFilter(self, a0: QObject | None, a1: QEvent | None):
        if a0 == self:
            a1 = cast(QEvent, a1)
            if a1.type() == QEvent.Type.FocusIn:
                self.setStyleSheet(self.focused_sheet)
            elif a1.type() == QEvent.Type.FocusOut:
                self.setStyleSheet(self.style_sheet_buffer)
            if (
                isinstance(a1, QKeyEvent)
                and a1.type() == a1.Type.KeyPress
                and (
                    a1.key()
                    in (
                        Qt.Key.Key_Tab,
                        Qt.Key.Key_Up,
                        Qt.Key.Key_Down,
                        Qt.Key.Key_Left,
                        Qt.Key.Key_Right,
                    )
                )
            ):
                # It doesn't work without the QTimer for some reason, probably cause the horizontal scroll bar centering bug happens
                # after this event is processed so the fix is overwridden hence we wait for the bug to happen first then fix it thus we need the QTimer
                QTimer.singleShot(0, self.search_window.fix_hor_scroll_bar)
        return super().eventFilter(a0, a1)


class SearchThread(QThread):
    finished = pyqtSignal(str, list)

    def __init__(self, search_window: SearchWindow, anime_title: str, site: str):
        super().__init__(search_window)
        self.anime_title = anime_title
        self.site = site

    def run(self):
        extracted_results = []
        if self.site == PAHE:
            results = pahe.search(self.anime_title)

            for result in results:
                title, page_link, anime_id = pahe.extract_anime_title_page_link_and_id(
                    result
                )
                extracted_results.append(Anime(title, page_link, anime_id))
        elif self.site == GOGO:
            results = gogo.search(self.anime_title)
            for title, page_link in results:
                extracted_results.append(Anime(title, page_link, None))
        self.finished.emit(self.site, extracted_results)



import os
import re
import subprocess
import sys
from base64 import b64decode
from string import ascii_letters, digits, printable
from threading import Event
from time import sleep as timesleep
from typing import Any, Callable, Iterator, Sequence, cast
from random import choice as random_choice
from webbrowser import open_new_tab

import requests

from senpwai.utils.static import log_exception, try_deleting

PARSER = "html.parser"
IBYTES_TO_MBS_DIVISOR = 1024 * 1024
QUALITY_REGEX_1 = re.compile(r"\b(\d{3,4})p\b")
QUALITY_REGEX_2 = re.compile(r"\b\d+x(\d+)\b")
NETWORK_RETRY_WAIT_TIME = 5
GITHUB_API_README_URL = "https://api.github.com/repos/SenZmaKi/Senpwai/readme"
RESOURCE_MOVED_STATUS_CODES = (301, 302, 307, 308)

FFMPEG_WINDOWS_INSTALLATION_GUIDE = "https://www.hostinger.com/tutorials/how-to-install-ffmpeg#How_to_Install_FFmpeg_on_Windows"
FFMPEG_LINUX_INSTALLATION_GUIDE = "https://www.hostinger.com/tutorials/how-to-install-ffmpeg#How_to_Install_FFmpeg_on_Linux"
FFMPEG_MAC_INSTALLATION_GUIDE = "https://www.hostinger.com/tutorials/how-to-install-ffmpeg#How_to_Install_FFmpeg_on_macOS"
USER_AGENTS = (
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/117.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/117.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.6 Safari/605.1.15",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:109.0) Gecko/20100101 Firefox/117.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/118.0",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/117.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:109.0) Gecko/20100101 Firefox/117.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36 Edg/116.0.1938.69",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36 OPR/102.0.0.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/115.0.0.0 Safari/537.36 OPR/101.0.0.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36 Edg/116.0.1938.76",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/115.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36 Edg/117.0.2045.31",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/115.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36 Edg/116.0.1938.81",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36 Edg/117.0.2045.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/115.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/116.0",
    "Mozilla/5.0 (X11; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/118.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.0 Safari/605.1.15",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36 Edg/117.0.2045.43",
    "Mozilla/5.0 (X11; Linux x86_64; rv:102.0) Gecko/20100101 Firefox/102.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:109.0) Gecko/20100101 Firefox/118.0",
    "Mozilla/5.0 (X11; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/116.0",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.75 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/115.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:109.0) Gecko/20100101 Firefox/115.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:102.0) Gecko/20100101 Firefox/102.0",
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/116.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:102.0) Gecko/20100101 Firefox/102.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36 Edg/116.0.1938.62",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/115.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.5.2 Safari/605.1.15",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36 Edg/117.0.2045.41",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:109.0) Gecko/20100101 Firefox/116.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:109.0) Gecko/20100101 Firefox/118.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
    "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.4 Safari/605.1.15",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.5 Safari/605.1.15",
)


def get_new_home_url_from_readme(site_name: str) -> str:
    """
    If the either Animepahe's or Gogoanime's domain name changes be sure to update it in the readme, specifically in the hyperlinks i.e., [Animepahe](https://animepahe.ru)
    and without an ending "/" i.e., https://animepahe.ru instead of https://animepahe.ru/ Also test if Senpwai is properly extracting it incase you made a mistake.

    :param site_name: Can be either Animepahe or Gogoanime.
    :return: The new home url.
    """
    encoded_readme_text = CLIENT.get(GITHUB_API_README_URL).json()["content"]
    readme_text = b64decode(encoded_readme_text).decode("utf-8")
    new_domain_name = cast(
        re.Match, re.search(rf"\[{site_name}\]\((.*)\)", readme_text)
    ).group(1)
    return new_domain_name


class DomainNameError(Exception):
    message = "The domain might have changed"

    def __init__(self, original_exception: Exception) -> None:
        super().__init__(DomainNameError.message, original_exception)
        self.original_exception = original_exception


def has_valid_internet_connection() -> bool:
    try:
        requests.get("https://www.google.com")
        return True
    except requests.exceptions.RequestException:
        return False


class Client:
    def __init__(self) -> None:
        self.headers = self.setup_request_headers()

    def setup_request_headers(self) -> dict[str, str]:
        headers = {"User-Agent": random_choice(USER_AGENTS)}
        return headers

    def append_headers(self, to_append: dict) -> dict:
        to_append.update(self.headers)
        return to_append

    def make_request(
        self,
        method: str,
        url: str,
        headers: dict | None,
        cookies={},
        stream=False,
        data: dict | bytes | None = None,
        json: dict | None = None,
        allow_redirects=False,
        timeout: int | None = None,
        exceptions_to_ignore: Sequence[type[Exception]] | None = None,
    ) -> requests.Response:
        if not headers:
            headers = self.headers
        if method == "GET":

            def callback():
                return requests.get(
                    url,
                    headers=headers,
                    stream=stream,
                    cookies=cookies,
                    allow_redirects=allow_redirects,
                    timeout=timeout,
                )
        else:

            def callback():
                return requests.post(
                    url,
                    headers=headers,
                    cookies=cookies,
                    data=data,
                    json=json,
                    allow_redirects=allow_redirects,
                )

        return cast(
            requests.Response,
            self.network_error_retry_wrapper(callback, exceptions_to_ignore),
        )

    def get(
        self,
        url: str,
        stream=False,
        headers: dict | None = None,
        timeout: int | None = None,
        cookies={},
        exceptions_to_ignore: Sequence[type[Exception]] | None = None,
    ) -> requests.Response:
        return self.make_request(
            "GET",
            url,
            headers,
            stream=stream,
            timeout=timeout,
            cookies=cookies,
            exceptions_to_ignore=exceptions_to_ignore,
        )

    def post(
        self,
        url: str,
        data: dict | bytes | None = None,
        json: dict | None = None,
        headers: dict | None = None,
        cookies={},
        allow_redirects=False,
        exceptions_to_ignore: Sequence[type[Exception]] | None = None,
    ) -> requests.Response:
        return self.make_request(
            "POST",
            url,
            headers,
            data=data,
            json=json,
            cookies=cookies,
            allow_redirects=allow_redirects,
            exceptions_to_ignore=exceptions_to_ignore,
        )

    def network_error_retry_wrapper(
        self,
        callback: Callable[[], Any],
        exceptions_to_ignore: Sequence[type[Exception]] | None = None,
    ) -> Any:
        while True:
            try:
                return callback()
            except requests.exceptions.RequestException as e:
                if (
                    exceptions_to_ignore is not None
                    and DomainNameError in exceptions_to_ignore
                ):
                    e = DomainNameError(e) if has_valid_internet_connection() else e
                if exceptions_to_ignore is not None and any(
                    [isinstance(e, exception) for exception in exceptions_to_ignore]
                ):
                    raise e
                log_exception(e)
                timesleep(1)


CLIENT = Client()


class AnimeMetadata:
    def __init__(
        self,
        poster_url: str,
        summary: str,
        episode_count: int,
        status: str,
        genres: list[str],
        release_year: int,
    ):
        self.poster_url = poster_url
        self.summary = summary
        self.episode_count = episode_count
        self.airing_status = status
        self.genres = genres
        self.release_year = release_year

    def get_poster_bytes(self) -> bytes:
        response = CLIENT.get(self.poster_url)
        return response.content


def match_quality(potential_qualities: list[str], user_quality: str) -> int:
    detected_qualities: list[tuple[int, int]] = []
    user_quality = user_quality.replace("p", "")
    for idx, potential_quality in enumerate(potential_qualities):
        match = QUALITY_REGEX_1.search(potential_quality)
        if not match:
            match = QUALITY_REGEX_2.search(potential_quality)

        if match:
            quality = cast(str, match.group(1))
            if quality == user_quality:
                return idx
            else:
                detected_qualities.append((int(quality), idx))
    int_user_quality = int(user_quality)
    if not detected_qualities:
        if int_user_quality <= 480:
            return 0
        return -1

    detected_qualities.sort(key=lambda x: x[0])
    closest = detected_qualities[0]
    for quality in detected_qualities:
        if quality[0] > int_user_quality:
            break
        closest = quality
    return closest[1]


def run_process_silently(args: list[str]) -> subprocess.CompletedProcess[bytes]:
    if sys.platform == "win32":
        return subprocess.run(args, creationflags=subprocess.CREATE_NO_WINDOW)
    return subprocess.run(args)


def run_process_in_new_console(
    args: list[str] | str,
) -> subprocess.CompletedProcess[bytes]:
    if sys.platform == "win32":
        return subprocess.run(args, creationflags=subprocess.CREATE_NEW_CONSOLE)
    return subprocess.run(args, shell=True)


def try_installing_ffmpeg() -> bool:
    if sys.platform == "win32":
        try:
            run_process_in_new_console("winget install Gyan.FFmpeg")
        # I should probably catch the specific exceptions but I'm too lazy to figure out all the possible exceptions
        except Exception as _:
            pass
        # Incase the installation was scuffed
        if ffmpeg_is_installed():
            return True
        else:
            open_new_tab(FFMPEG_WINDOWS_INSTALLATION_GUIDE)
            return False
    elif sys.platform == "linux":
        try:
            run_process_in_new_console(
                "sudo apt-get update && sudo apt-get install ffmpeg"
            )
        except Exception as _:
            pass
        if ffmpeg_is_installed():
            return True
        else:
            open_new_tab(FFMPEG_LINUX_INSTALLATION_GUIDE)
            return False

    else:
        try:
            run_process_in_new_console("brew install ffmpeg")
        except Exception as _:
            pass
        if ffmpeg_is_installed():
            return True
        else:
            open_new_tab(FFMPEG_MAC_INSTALLATION_GUIDE)
            return False


def fuzz_str(text: str) -> str:
    return sanitise_title(text, True).lower()


def sanitise_title(title: str, all=False, exclude="") -> str:
    if all:
        allowed_chars = set(ascii_letters + digits + exclude)
    else:
        allowed_chars = set(printable) - set('\\/:*?"<>|')
        title = title.replace(":", " -")
    sanitised = "".join([char for char in title if char in allowed_chars])

    return sanitised[:255].rstrip()


def lacked_episode_numbers(
    start_episode: int, end_episode: int, haved_episodes: list[int]
) -> list[int]:
    predicted_episodes_to_download: list[int] = []
    for episode in range(start_episode, end_episode + 1):
        if episode not in haved_episodes:
            predicted_episodes_to_download.append(episode)
    return predicted_episodes_to_download


def lacked_episodes(
    lacking_episode_numbers: list[int], episode_page_links: list[str]
) -> list[str]:
    # Episode count is what is used to generate lacking_episode_numbers, episode count is gotten from anime the page which uses subbed episodes count
    # so for an anime where sub episodes are ahead of dub the missing episodes will be outside the range of the episode_page_links
    first_eps_number = lacking_episode_numbers[0]
    return [
        episode_page_links[eps_number - first_eps_number]
        for eps_number in lacking_episode_numbers
    ]


def ffmpeg_is_installed() -> bool:
    try:
        run_process_silently(["ffmpeg"])
        return True
    except FileNotFoundError:
        return False


class ProgressFunction:
    def __init__(self) -> None:
        self.resume = Event()
        self.resume.set()
        self.cancelled = False

    def pause_or_resume(self):
        if self.resume.is_set():
            return self.resume.clear()
        self.resume.set()

    def cancel(self):
        if self.resume.is_set():
            self.cancelled = True


class Download(ProgressFunction):
    def __init__(
        self,
        link_or_segment_urls: str | list[str],
        episode_title: str,
        download_folder_path: str,
        progress_update_callback: Callable = lambda _: None,
        file_extension=".mp4",
        is_hls_download=False,
        cookies=requests.sessions.RequestsCookieJar(),
    ) -> None:
        super().__init__()
        self.link_or_segment_urls = link_or_segment_urls
        self.episode_title = episode_title
        self.download_folder_path = download_folder_path
        self.progress_update_callback = progress_update_callback
        self.is_hls_download = is_hls_download
        self.cookies = cookies
        file_title = f"{self.episode_title}{file_extension}"
        self.file_path = os.path.join(self.download_folder_path, file_title)
        ext = ".ts" if is_hls_download else file_extension
        temporary_file_title = f"{self.episode_title} [Downloading]{ext}"
        self.temporary_file_path = os.path.join(
            self.download_folder_path, temporary_file_title
        )
        if os.path.isfile(self.temporary_file_path):
            try_deleting(self.temporary_file_path)

    @staticmethod
    def get_resource_length(url: str) -> tuple[int, str]:
        response = CLIENT.get(url, stream=True)
        new_location = response.headers.get("Location", "")
        if not new_location and response.status_code in RESOURCE_MOVED_STATUS_CODES:
            return Download.get_resource_length(new_location)
        return (int(response.headers.get("Content-Length", 0)), url)

    def cancel(self):
        return super().cancel()

    def start_download(self):
        download_complete = False
        while not download_complete and not self.cancelled:
            if self.is_hls_download:
                download_complete = self.hls_download()
            else:
                download_complete = self.normal_download()
        if self.cancelled:
            try_deleting(self.temporary_file_path)
            return
        try_deleting(self.file_path)
        if self.is_hls_download:
            run_process_silently(
                ["ffmpeg", "-i", self.temporary_file_path, "-c", "copy", self.file_path]
            )
            return try_deleting(self.temporary_file_path)
        try:
            return os.rename(self.temporary_file_path, self.file_path)
        except PermissionError:  # Maybe they started watching the episode on VLC before it finished downloading now VLC has a handle to the file hence PermissionDenied
            pass

    def hls_download(self) -> bool:
        with open(self.temporary_file_path, "wb") as f:
            for seg in self.link_or_segment_urls:
                response = CLIENT.get(seg)
                self.resume.wait()
                if self.cancelled:
                    return False
                f.write(response.content)
                self.progress_update_callback(1)
            return True

    def normal_download(self) -> bool:
        self.link_or_segment_urls = cast(str, self.link_or_segment_urls)
        response = CLIENT.get(
            self.link_or_segment_urls, stream=True, timeout=30, cookies=self.cookies
        )

        def response_ranged(start_byte):
            self.link_or_segment_urls = cast(str, self.link_or_segment_urls)
            return CLIENT.get(
                self.link_or_segment_urls,
                stream=True,
                headers=CLIENT.append_headers({"Range": f"bytes={start_byte}-"}),
                timeout=30,
                cookies=self.cookies,
            )

        total = int(response.headers.get("Content-Length", 0))

        def download(start_byte: int = 0) -> bool:
            mode = "wb" if start_byte == 0 else "ab"
            with open(self.temporary_file_path, mode) as file:
                iter_content = cast(
                    Iterator[bytes],
                    response.iter_content(chunk_size=IBYTES_TO_MBS_DIVISOR)
                    if start_byte == 0
                    else CLIENT.network_error_retry_wrapper(
                        lambda: response_ranged(start_byte).iter_content(
                            chunk_size=IBYTES_TO_MBS_DIVISOR
                        )
                    ),
                )
                while True:
                    try:

                        def get_data():
                            return next(iter_content)

                        data = cast(bytes, CLIENT.network_error_retry_wrapper(get_data))
                        self.resume.wait()
                        if self.cancelled:
                            return False
                        size = file.write(data)
                        self.progress_update_callback(size)
                    except StopIteration:
                        break

            file_size = os.path.getsize(self.temporary_file_path)
            return True if file_size >= total else download(file_size)

        return download()


if __name__ == "__main__":
    pass



import re
PAHE = "pahe"
PAHE_DOMAIN = "animepahe.ru"
PAHE_HOME_URL = f"https://{PAHE_DOMAIN}"
FULL_SITE_NAME = "Animepahe"
API_ENTRY_POINT = f"{PAHE_HOME_URL}/api?m="
"""
Base URL for Pahe's API, used for retrieving anime information.
"""
ANIME_PAGE_URL = f"{API_ENTRY_POINT}release&id={{}}&sort=episode_asc"
"""
Generates the anime page link from the provided anime id
Example: https://animepahe.ru/api?m=release&id={anime_id}&sort=episode_asc
"""
EPISODE_PAGE_URL = f"{PAHE_HOME_URL}/play/{{}}/{{}}"
"""
Generates episode page link from the provided anime id and episode session
Example: https://animepahe.ru/play/{anime_id}/{episode_session}
"""
LOAD_EPISODES_URL = "{}&page={}"
"""
Generates the load episodes link from the provided anime page link and page number.
Example: {anime_page_link}&page={page_number}
"""
KWIK_DOMAIN = "kwik.si"
KWIK_PAGE_REGEX = re.compile(rf"https?://{KWIK_DOMAIN}/f/([^\"']+)")
DUB_PATTERN = "eng"
EPISODE_SIZE_REGEX = re.compile(r"\b(\d+)MB\b")
PARAM_REGEX = re.compile(r"""\(\"(\w+)\",\d+,\"(\w+)\",(\d+),(\d+),(\d+)\)""")




import json
import os
from pathlib import Path
from typing import cast

import anitopy
from appdirs import user_config_dir
from senpwai.scrapers import gogo, pahe
import logging
import re

from senpwai.utils.scraper import (
    CLIENT,
    AnimeMetadata,
    sanitise_title,
)
from senpwai.utils.static import (
    APP_NAME,
    generate_windows_setup_file_titles,
    VERSION,
    GOGO,
    GOGO_HLS_MODE,
    GOGO_NORM_MODE,
    PAHE,
    Q_720,
    SUB,
)

VERSION_REGEX = re.compile(r"(\d+(\.\d+)*)")


def update_available(
    latest_release_api_url: str, app_name: str, curr_version: str
) -> tuple[bool, str, str, str]:
    latest_version_json = CLIENT.get(latest_release_api_url).json()
    latest_version_tag = latest_version_json["tag_name"]
    match = cast(re.Match, VERSION_REGEX.search(latest_version_tag))
    latest_version = match.group(1)
    download_url = ""
    target_asset_name = ""
    # Update this logic if u ever officially release on Linux or Mac
    target_asset_names = generate_windows_setup_file_titles(app_name)
    update_available = True if latest_version != curr_version else False
    if update_available:
        for asset in latest_version_json["assets"]:
            matched_asset = None
            for tan in target_asset_names:
                if asset["name"] == tan:
                    matched_asset = tan
                    break
            if matched_asset is not None:
                download_url, target_asset_name = (
                    asset["browser_download_url"],
                    matched_asset,
                )
                break
    return (
        update_available,
        download_url,
        target_asset_name,
        latest_version_json["body"],
    )


class Settings:
    types = str | int | bool | list[str]

    def __init__(self) -> None:
        self.config_dir = self.setup_config_dir()
        self.settings_json_path = os.path.join(self.config_dir, "settings.json")
        # Default settings
        # Only these settings will be saved to settings.json
        # Everytime you add a new class member that isn't a setting, make sure to pop it from the Settings.dict_settings method
        self.sub_or_dub = SUB
        self.quality = Q_720
        self.download_folder_paths = self.setup_default_download_folder()
        self.max_simultaneous_downloads = 2
        self.allow_notifications = True
        self.start_in_fullscreen = True
        self.run_on_startup = False
        self.gogo_norm_or_hls_mode = GOGO_NORM_MODE
        self.tracked_anime: list[str] = []
        self.auto_download_site = PAHE
        self.check_for_new_eps_after = 24
        self.gogo_skip_calculate = False
        self.version = VERSION

        self.configure_settings()
        self.setup_logger()
        self.is_update_install = self.version != VERSION

    def setup_logger(self) -> None:
        error_logs_file_path = os.path.join(self.config_dir, "errors.log")
        if not os.path.exists(error_logs_file_path):
            f = open(error_logs_file_path, "w")
            f.close()

        logging.basicConfig(
            filename=error_logs_file_path,
            level=logging.ERROR,
            format="%(asctime)s - %(levelname)s - %(message)s",
        )

    def setup_config_dir(self) -> str:
        config_dir = os.path.join(user_config_dir(), APP_NAME)
        if not os.path.isdir(config_dir):
            os.makedirs(config_dir)
        return config_dir

    def configure_settings(self) -> None:
        failed_to_load_settings = (
            False if os.path.isfile(self.settings_json_path) else True
        )
        if not failed_to_load_settings:
            with open(self.settings_json_path, "r") as f:
                try:
                    settings = cast(dict, json.load(f))
                    try:
                        self.__dict__.update(settings)
                    except ValueError:
                        failed_to_load_settings = True
                except json.decoder.JSONDecodeError:
                    failed_to_load_settings = True
        if failed_to_load_settings:
            with open(self.settings_json_path, "w") as f:
                json.dump(self.dict_settings(), f, indent=4)

    def setup_default_download_folder(self) -> list[str]:
        downloads_folder = os.path.join(Path.home(), "Downloads", "Anime")
        if not os.path.isfile(self.settings_json_path) and not os.path.isdir(
            downloads_folder
        ):
            os.makedirs(downloads_folder)
        return [downloads_folder]

    def dict_settings(self) -> dict:
        d_settings = {k: v for k, v in self.__dict__.items()}
        d_settings.pop("settings_json_path")
        d_settings.pop("config_dir")
        return d_settings

    def update_sub_or_dub(self, sub_or_dub: str) -> None:
        self.sub_or_dub = sub_or_dub
        self.save_settings()

    def update_quality(self, quality: str) -> None:
        self.quality = quality
        self.save_settings()

    def update_download_folder_paths(self, download_folder_paths: list[str]) -> None:
        self.download_folder_paths = download_folder_paths
        self.save_settings()

    def add_download_folder_path(self, download_folder_path: str) -> None:
        self.download_folder_paths.append(download_folder_path)
        self.save_settings()

    def remove_download_folder_path(self, download_folder_path: str) -> None:
        self.download_folder_paths.remove(download_folder_path)
        self.save_settings()

    def change_download_folder_path(self, old_path_idx: int, new_path: str) -> None:
        self.download_folder_paths[old_path_idx] = new_path
        self.save_settings()

    def pop_download_folder_path(self, index: int) -> None:
        self.download_folder_paths.pop(index)
        self.save_settings()

    def update_max_simultaneous_downloads(
        self, max_simultaneous_downloads: int
    ) -> None:
        self.max_simultaneous_downloads = max_simultaneous_downloads
        self.save_settings()

    def update_allow_notifications(self, allow_notifications: bool) -> None:
        self.allow_notifications = allow_notifications
        self.save_settings()

    def update_start_in_fullscreen(self, start_in_fullscreen: bool) -> None:
        self.start_in_fullscreen = start_in_fullscreen
        self.save_settings()

    def update_run_on_startup(self, run_on_startup: bool) -> None:
        self.run_on_startup = run_on_startup
        self.save_settings()

    def update_gogo_norm_or_hls_mode(self, gogo_norm_or_hls_mode: str) -> None:
        self.gogo_norm_or_hls_mode = gogo_norm_or_hls_mode
        self.save_settings()

    def update_tracked_anime(self, tracked_anime: list[str]) -> None:
        self.tracked_anime = tracked_anime
        self.save_settings()

    def remove_tracked_anime(self, anime_name: str) -> None:
        self.tracked_anime.remove(anime_name)
        self.save_settings()

    def add_tracked_anime(self, anime_name: str) -> None:
        self.tracked_anime.append(anime_name)
        self.save_settings()

    def update_auto_download_site(self, auto_download_site: str) -> None:
        self.auto_download_site = auto_download_site
        self.save_settings()

    def update_check_for_new_eps_after(self, check_for_new_eps_after: int) -> None:
        self.check_for_new_eps_after = check_for_new_eps_after
        self.save_settings()

    def update_gogo_skip_calculate(self, gogo_skip_calculate: bool) -> None:
        self.gogo_skip_calculate = gogo_skip_calculate
        self.save_settings()

    def update_pahe_home_url(self, pahe_home_url: str) -> None:
        self.pahe_home_url = pahe_home_url
        self.save_settings()

    def update_gogo_home_url(self, gogo_home_url: str) -> None:
        self.gogo_home_url = gogo_home_url
        self.save_settings()

    def save_settings(self) -> None:
        with open(self.settings_json_path, "w") as f:
            json.dump(self.dict_settings(), f, indent=4)


SETTINGS = Settings()


class Anime:
    def __init__(self, title: str, page_link: str, anime_id: str | None) -> None:
        self.title = title
        self.page_link = page_link
        self.id = anime_id


class AnimeDetails:
    def __init__(self, anime: Anime, site: str):
        self.anime = anime
        self.site = site
        self.is_hls_download = (
            True
            if site == GOGO and SETTINGS.gogo_norm_or_hls_mode == GOGO_HLS_MODE
            else False
        )
        self.sanitised_title = sanitise_title(anime.title)
        self.default_download_path = SETTINGS.download_folder_paths[0]
        self.anime_folder_path = self.get_anime_folder_path()
        self.potentially_haved_episodes = list(Path(self.anime_folder_path).glob("*"))
        self.haved_episodes: list[int] = []
        (
            self.haved_start,
            self.haved_end,
            self.haved_count,
        ) = self.get_start_end_and_count_of_haved_episodes()
        self.dub_available, self.dub_page_link = self.get_dub_availablilty_status()
        self.metadata = self.get_metadata()
        self.episode_count = self.metadata.episode_count
        self.quality = SETTINGS.quality
        self.sub_or_dub = SETTINGS.sub_or_dub
        self.ddls_or_segs_urls: list[str] | list[list[str]] = []
        self.download_info: list[str] = []
        self.total_download_size: int = 0
        self.lacked_episode_numbers: list[int] = []
        self.skip_calculating_size = (
            True
            if site == GOGO
            and not self.is_hls_download
            and SETTINGS.gogo_skip_calculate
            else False
        )

    def episode_title(self, lacked_eps_idx: int) -> str:
        episode_number_str = str(self.lacked_episode_numbers[lacked_eps_idx]).zfill(2)
        return f"{self.sanitised_title} E{episode_number_str}"

    def validate_anime_folder_path(self) -> None:
        if not os.path.isdir(self.anime_folder_path):
            os.makedirs(self.anime_folder_path)

    def get_anime_folder_path(self) -> str:
        def try_path(title: str) -> str | None:
            for path in SETTINGS.download_folder_paths:
                potential = os.path.join(path, title)
                if os.path.isdir(potential):
                    return potential
            return None

        fully_sanitised_title = sanitise_title(self.anime.title, True, " ")
        parent_seasons_path = ""
        season_number = 1
        parsed_title = ""
        anime_type = ""
        parsed = {}

        def init(title: str):
            nonlocal \
                parsed, \
                parsed_title, \
                parent_seasons_path, \
                anime_type, \
                season_number
            parsed = anitopy.parse(title)
            if parsed:
                parsed_title = parsed.get("anime_title", title)
                # It could be that the anime is a Special/OVA/ONA
                anime_type = parsed.get("anime_type", "")
                if anime_type:
                    # In the resulting parsed anime_title, Anitopy only ignores Seasons but not Types for some reason, e.g., "Attack On Titan Season 1" will
                    # be parsed to "Attack on Titan" meanwhile, "Attack on Titan Specials" will still remain as "Attack on Titan"
                    parsed_title = parsed_title.replace(anime_type, "").strip()
                parent_seasons_path = try_path(parsed_title)
                if not anime_type:
                    season_number = parsed.get("anime_season", 1)

        init(self.sanitised_title)
        if not parent_seasons_path:
            init(fully_sanitised_title)
        if parent_seasons_path and parsed_title and parsed:
            target_folders = (
                [anime_type]
                if anime_type
                else [
                    f"Season {season_number}",
                    f"SN {season_number}",
                    f"Sn {season_number}",
                    f"{parsed_title} Season {season_number}",
                    f"{parsed_title} SN {season_number}",
                    f"{parsed_title} Sn {season_number}",
                ]
            )
            target_folders += [self.sanitised_title, fully_sanitised_title]

            for f in target_folders:
                folder = os.path.join(parent_seasons_path, f)
                if os.path.isdir(folder):
                    return folder
        if path := try_path(self.sanitised_title):
            return path
        elif path := try_path(fully_sanitised_title):
            return path
        else:
            return os.path.join(self.default_download_path, self.sanitised_title)

    def get_start_end_and_count_of_haved_episodes(
        self,
    ) -> tuple[int, int, int] | tuple[None, None, None]:
        if self.potentially_haved_episodes:
            for episode in self.potentially_haved_episodes:
                if "[Downloading]" in episode.name:
                    continue
                parsed = anitopy.parse(episode.name)
                if not parsed:
                    continue
                try:
                    episode_number = int(parsed["episode_number"])
                except (KeyError, ValueError):
                    continue
                if episode_number > 0:
                    self.haved_episodes.append(episode_number)
            self.haved_episodes.sort()
        return (
            (self.haved_episodes[0], self.haved_episodes[-1], len(self.haved_episodes))
            if self.haved_episodes
            else (None, None, None)
        )

    def get_dub_availablilty_status(self) -> tuple[bool, str]:
        if self.site == PAHE:
            dub_available = pahe.dub_available(
                self.anime.page_link, cast(str, self.anime.id)
            )
            link = self.anime.page_link
        else:
            dub_available, link = gogo.dub_availability_and_link(self.anime.title)
        return dub_available, link

    def get_metadata(self) -> AnimeMetadata:
        if self.site == PAHE:
            metadata = pahe.get_anime_metadata(cast(str, self.anime.id))
        else:
            page_content = gogo.get_anime_page_content(self.anime.page_link)
            metadata = gogo.extract_anime_metadata(page_content)
        return metadata



import os
from typing import TYPE_CHECKING, Callable, cast
import sys
from pylnk3 import for_file as pylnk3_for_file
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QFileDialog, QHBoxLayout, QLayoutItem, QVBoxLayout, QWidget
from senpwai.utils.classes import SETTINGS
from senpwai.utils.scraper import try_deleting
from senpwai.utils.static import (
    AMOGUS_EASTER_EGG,
    APP_EXE_PATH,
    APP_EXE_ROOT_DIRECTORY,
    APP_NAME,
    DUB,
    GOGO,
    GOGO_HLS_MODE,
    GOGO_NORM_MODE,
    GOGO_NORMAL_COLOR,
    GOGO_PRESSED_COLOR,
    IS_PIP_INSTALL,
    MINIMISED_TO_TRAY_ARG,
    PAHE,
    PAHE_HOVER_COLOR,
    PAHE_NORMAL_COLOR,
    PAHE_PRESSED_COLOR,
    Q_360,
    Q_480,
    Q_720,
    Q_1080,
    RED_HOVER_COLOR,
    RED_NORMAL_COLOR,
    RED_PRESSED_COLOR,
    SETTINGS_WINDOW_BCKG_IMAGE_PATH,
    SUB,
    requires_admin_access,
)
from senpwai.utils.widgets import (
    ErrorLabel,
    GogoNormOrHlsButton,
    HorizontalLine,
    NumberInput,
    OptionButton,
    QualityButton,
    ScrollableSection,
    StyledButton,
    StyledLabel,
    SubDubButton,
    fix_qt_path_for_windows,
    set_minimum_size_policy,
)

from senpwai.windows.download import DownloadWindow
from senpwai.windows.main import AbstractWindow
# https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports/3957388#39757388
if TYPE_CHECKING:
    from senpwai.windows.main import MainWindow

class SettingsWindow(AbstractWindow):
    def __init__(self, main_window: 'MainWindow') -> None:
        super().__init__(main_window, SETTINGS_WINDOW_BCKG_IMAGE_PATH)
        self.font_size = 15
        main_layout = QHBoxLayout()
        main_widget = ScrollableSection(main_layout)
        left_layout = QVBoxLayout()
        left_widget = QWidget()
        left_widget.setLayout(left_layout)
        right_layout = QVBoxLayout()
        right_widget = QWidget()
        right_widget.setLayout(right_layout)
        main_layout.addWidget(left_widget)
        main_layout.addWidget(right_widget)

        self.sub_dub_setting = SubDubSetting(self)
        self.quality_setting = QualitySetting(self)
        self.max_simultaneous_downloads_setting = MaxSimultaneousDownloadsSetting(self)
        self.make_download_complete_notification_setting = AllowNotificationsSetting(
            self
        )
        self.start_in_fullscreen = StartInFullscreenSetting(self)
        self.download_folder_setting = DownloadFoldersSetting(self, main_window)
        self.gogo_norm_or_hls_mode_setting = GogoNormOrHlsSetting(self)
        self.tracked_anime = TrackedAnimeListSetting(
            self,
        )
        self.auto_download_site = AutoDownloadSite(self)
        self.check_for_new_eps_after = CheckForNewEpsAfterSetting(
            self, main_window.download_window
        )
        self.gogo_skip_calculate = GogoSkipCalculate(self)
        left_layout.addWidget(self.sub_dub_setting)
        left_layout.addWidget(self.quality_setting)
        left_layout.addWidget(self.max_simultaneous_downloads_setting)
        left_layout.addWidget(self.gogo_norm_or_hls_mode_setting)
        left_layout.addWidget(self.gogo_skip_calculate)
        left_layout.addWidget(self.make_download_complete_notification_setting)
        left_layout.addWidget(self.start_in_fullscreen)
        if sys.platform == "win32" and not IS_PIP_INSTALL and APP_EXE_PATH:
            self.run_on_startup = RunOnStartUp(self)
            left_layout.addWidget(self.run_on_startup)
        right_layout.addWidget(self.download_folder_setting)
        right_layout.addWidget(self.auto_download_site)
        right_layout.addWidget(self.check_for_new_eps_after)
        right_layout.addWidget(self.tracked_anime)
        self.full_layout.addWidget(main_widget)
        self.setLayout(self.full_layout)


class FolderSetting(QWidget):
    def __init__(
        self,
        settings_window: SettingsWindow,
        main_window: 'MainWindow',
        setting_info: str,
        setting_tool_tip: str | None,
    ):
        super().__init__()
        self.settings_window = settings_window
        self.font_size = settings_window.font_size
        self.main_window = main_window
        self.main_layout = QVBoxLayout()
        settings_label = StyledLabel(font_size=self.font_size + 5)
        settings_label.setText(setting_info)
        if setting_tool_tip:
            settings_label.setToolTip(setting_tool_tip)
        set_minimum_size_policy(settings_label)
        self.error_label = ErrorLabel(18, 6)
        self.error_label.hide()
        add_button = StyledButton(
            self,
            self.font_size,
            "white",
            "green",
            GOGO_NORMAL_COLOR,
            GOGO_PRESSED_COLOR,
        )
        add_button.clicked.connect(self.add_folder_to_settings)
        add_button.setText("ADD")
        set_minimum_size_policy(add_button)
        settings_label_and_add_button_widget = QWidget()
        settings_label_and_add_button_layout = QHBoxLayout()
        settings_label_and_add_button_layout.addWidget(settings_label)
        settings_label_and_add_button_layout.addWidget(add_button)
        settings_label_and_add_button_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        settings_label_and_add_button_widget.setLayout(
            settings_label_and_add_button_layout
        )
        self.main_layout.addWidget(self.error_label)
        self.main_layout.addWidget(settings_label_and_add_button_widget)
        line = HorizontalLine()
        line.setFixedHeight(7)
        self.main_layout.addWidget(line)
        self.folder_widgets_layout = QVBoxLayout()
        for idx, folder in enumerate(SETTINGS.download_folder_paths):
            self.folder_widgets_layout.addWidget(
                FolderWidget(main_window, self, 14, folder, idx),
                alignment=Qt.AlignmentFlag.AlignTop,
            )
        folder_widgets_widget = ScrollableSection(self.folder_widgets_layout)
        self.main_layout.addWidget(folder_widgets_widget)
        self.setLayout(self.main_layout)

    def error(self, error_message: str):
        self.error_label.setText(error_message)
        self.error_label.update()
        set_minimum_size_policy(self.error_label)
        self.error_label.show()

    def is_valid_new_folder(self, new_folder_path: str) -> bool:
        if requires_admin_access(new_folder_path):
            self.error("The folder you chose requires admin access so i've ignored it")
            return False
        elif not os.path.isdir(new_folder_path):
            self.error("Choose a valid folder, onegaishimasu")
            return False
        elif new_folder_path in SETTINGS.download_folder_paths:
            self.error("Baka!!! that folder is already in the settings")
            return False
        else:
            return True

    def update_widget_indices(self):
        for idx in range(self.folder_widgets_layout.count()):
            cast(
                FolderWidget,
                cast(QLayoutItem, self.folder_widgets_layout.itemAt(idx)).widget(),
            ).index = idx

    def change_from_folder_settings(self, new_folder_path: str, folder_widget: 'FolderWidget'):
        SETTINGS.change_download_folder_path(folder_widget.index, new_folder_path)
        folder_widget.folder_path = new_folder_path
        folder_widget.folder_label.setText(new_folder_path)
        set_minimum_size_policy(folder_widget.folder_label)
        folder_widget.folder_label.update()

    def remove_from_folder_settings(self, folder_widget: 'FolderWidget'):
        SETTINGS.pop_download_folder_path(folder_widget.index)
        folder_widget.deleteLater()
        self.update_widget_indices()

    def add_folder_to_settings(self):
        added_folder_path = QFileDialog.getExistingDirectory(
            self.main_window, "Choose folder"
        )
        added_folder_path = fix_qt_path_for_windows(added_folder_path)
        if not self.is_valid_new_folder(added_folder_path):
            return
        SETTINGS.add_download_folder_path(added_folder_path)
        self.folder_widgets_layout.addWidget(
            FolderWidget(
                self.main_window,
                self,
                14,
                added_folder_path,
                self.folder_widgets_layout.count(),
            ),
            alignment=Qt.AlignmentFlag.AlignTop,
        )


class DownloadFoldersSetting(FolderSetting):
    def __init__(self, settings_window: SettingsWindow, main_window: 'MainWindow'):
        super().__init__(
            settings_window,
            main_window,
            "Download folders",
            f"{APP_NAME} will search these folders for anime episodes, in the order shown",
        )

    def remove_from_folder_settings(self, folder_widget: 'FolderWidget'):
        if len(SETTINGS.download_folder_paths) == 1:
            return self.error("Yarou!!! You must have at least one download folder")
        return super().remove_from_folder_settings(folder_widget)


class FolderWidget(QWidget):
    def __init__(
        self,
        main_window: 'MainWindow',
        folder_setting: FolderSetting,
        font_size: int,
        folder_path: str,
        index: int,
    ):
        super().__init__()
        self.main_window = main_window
        self.folder_path = folder_path
        self.folder_setting = folder_setting
        self.index = index
        main_layout = QHBoxLayout()
        self.folder_label = StyledLabel(font_size=font_size)
        self.folder_label.setText(folder_path)
        set_minimum_size_policy(self.folder_label)
        self.change_button = StyledButton(
            self,
            font_size,
            "white",
            PAHE_NORMAL_COLOR,
            PAHE_HOVER_COLOR,
            PAHE_PRESSED_COLOR,
        )
        self.change_button.clicked.connect(self.change_folder)
        self.change_button.setText("CHANGE")
        set_minimum_size_policy(self.change_button)
        remove_button = StyledButton(
            self,
            font_size,
            "white",
            RED_NORMAL_COLOR,
            RED_HOVER_COLOR,
            RED_PRESSED_COLOR,
        )
        remove_button.setText("REMOVE")
        remove_button.clicked.connect(
            lambda: self.folder_setting.remove_from_folder_settings(self)
        )
        set_minimum_size_policy(remove_button)
        main_layout.addWidget(self.folder_label)
        main_layout.addWidget(self.change_button)
        main_layout.addWidget(remove_button)
        main_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.setLayout(main_layout)

    def change_folder(self):
        new_folder_path = QFileDialog.getExistingDirectory(
            self.main_window, "Choose folder"
        )
        new_folder_path = fix_qt_path_for_windows(new_folder_path)
        if self.folder_setting.is_valid_new_folder(new_folder_path):
            self.folder_setting.change_from_folder_settings(new_folder_path, self)


class YesOrNoButton(OptionButton):
    def __init__(self, yes_or_no: bool, font_size):
        super().__init__(
            None,
            yes_or_no,
            "YES" if yes_or_no else "NO",
            font_size,
            PAHE_NORMAL_COLOR,
            PAHE_PRESSED_COLOR,
        )


class SettingWidget(QWidget):
    def __init__(
        self,
        settings_window: SettingsWindow,
        setting_info: str,
        widgets_to_add: list,
        horizontal_layout=True,
    ):
        super().__init__()
        self.setting_label = StyledLabel(font_size=settings_window.font_size + 5)
        self.setting_label.setText(setting_info)
        set_minimum_size_policy(self.setting_label)
        if horizontal_layout:
            main_layout = QHBoxLayout()
        else:
            main_layout = QVBoxLayout()
        main_layout.addWidget(self.setting_label)
        for button in widgets_to_add:
            main_layout.addWidget(button)
        main_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.setLayout(main_layout)


class RemovableWidget(QWidget):
    def __init__(self, text: str, font_size: int = 20):
        super().__init__()
        main_layout = QHBoxLayout()
        self.text = text
        label = StyledLabel(self, font_size)
        label.setText(text)
        self.remove_button = StyledButton(
            self,
            font_size,
            "white",
            RED_NORMAL_COLOR,
            RED_HOVER_COLOR,
            RED_PRESSED_COLOR,
        )
        self.remove_button.setText("REMOVE")
        self.remove_button.clicked.connect(self.deleteLater)
        set_minimum_size_policy(self.remove_button)
        main_layout.addWidget(label)
        main_layout.addWidget(self.remove_button)
        self.setLayout(main_layout)


class TrackedAnimeListSetting(SettingWidget):
    def __init__(self, settings_window: SettingsWindow):
        self.main_layout = QVBoxLayout()
        self.settings_window = settings_window
        main_widget = ScrollableSection(self.main_layout)
        self.anime_buttons: list[RemovableWidget] = []
        for anime in SETTINGS.tracked_anime:
            wid = RemovableWidget(anime, font_size=14)
            self.setup_anime_widget(wid)
        line = HorizontalLine()
        line.setFixedHeight(7)
        super().__init__(
            settings_window,
            "Track for new episodes then auto download",
            [line, main_widget],
            False,
        )
        self.setting_label.setToolTip(
            "Senpwai will check for new episodes of your tracked anime when you start the app\nthen in intervals of the hours you specify so long as it is running"
        )

    def setup_anime_widget(self, wid: RemovableWidget):
        wid.remove_button.clicked.connect(
            lambda _, txt=wid.text: SETTINGS.remove_tracked_anime(txt)
        )
        wid.remove_button.clicked.connect(lambda: self.anime_buttons.remove(wid))
        set_minimum_size_policy(wid)
        self.main_layout.addWidget(wid, alignment=Qt.AlignmentFlag.AlignTop)
        self.anime_buttons.append(wid)

    def remove_anime(self, title: str):
        for anime in self.anime_buttons:
            if anime.text == title:
                return anime.remove_button.click()

    def add_anime(self, title: str):
        for anime in self.anime_buttons:
            if anime.text == title:
                return
        wid = RemovableWidget(title, 16)
        self.setup_anime_widget(wid)
        SETTINGS.add_tracked_anime(title)


class AutoDownloadSite(SettingWidget):
    def __init__(self, settings_window: SettingsWindow):
        self.font_size = settings_window.font_size
        pahe_button = OptionButton(
            None, PAHE, "PAHE", self.font_size, PAHE_NORMAL_COLOR, PAHE_PRESSED_COLOR
        )
        gogo_button = OptionButton(
            None, GOGO, "GOGO", self.font_size, GOGO_NORMAL_COLOR, GOGO_PRESSED_COLOR
        )
        pahe_button.clicked.connect(lambda: gogo_button.set_picked_status(False))
        pahe_button.clicked.connect(
            lambda: SETTINGS.update_auto_download_site(cast(str, pahe_button.option))
        )
        gogo_button.clicked.connect(lambda: pahe_button.set_picked_status(False))
        gogo_button.clicked.connect(
            lambda: SETTINGS.update_auto_download_site(cast(str, gogo_button.option))
        )
        if SETTINGS.auto_download_site == PAHE:
            pahe_button.set_picked_status(True)
        else:
            gogo_button.set_picked_status(True)

        super().__init__(
            settings_window, "Auto download site", [pahe_button, gogo_button]
        )
        self.setting_label.setToolTip(
            f"If {APP_NAME} can't find the anime in the specified site it will try the other"
        )


class YesOrNoSetting(SettingWidget):
    def __init__(
        self,
        settings_window: SettingsWindow,
        setting_info: str,
        tooltip: str | None = None,
    ):
        self.yes_button = YesOrNoButton(True, settings_window.font_size)
        self.no_button = YesOrNoButton(False, settings_window.font_size)
        self.yes_button.clicked.connect(lambda: self.no_button.set_picked_status(False))
        self.no_button.clicked.connect(lambda: self.yes_button.set_picked_status(False))
        set_minimum_size_policy(self.yes_button)
        set_minimum_size_policy(self.no_button)
        super().__init__(
            settings_window, setting_info, [self.yes_button, self.no_button]
        )

        if tooltip:
            self.setting_label.setToolTip(tooltip)


class GogoSkipCalculate(YesOrNoSetting):
    def __init__(self, settings_window: SettingsWindow):
        super().__init__(settings_window, "Skip calculating download size for Gogo")
        if SETTINGS.gogo_skip_calculate:
            self.yes_button.set_picked_status(True)
        else:
            self.no_button.set_picked_status(True)
        self.yes_button.clicked.connect(
            lambda: SETTINGS.update_gogo_skip_calculate(True)
        )
        self.no_button.clicked.connect(
            lambda: SETTINGS.update_gogo_skip_calculate(False)
        )


class StartInFullscreenSetting(YesOrNoSetting):
    def __init__(self, settings_window: SettingsWindow):
        super().__init__(settings_window, "Start app in fullscreen")
        if SETTINGS.start_in_fullscreen:
            self.yes_button.set_picked_status(True)
        else:
            self.no_button.set_picked_status(True)
        self.yes_button.clicked.connect(
            lambda: SETTINGS.update_start_in_fullscreen(True)
        )
        self.no_button.clicked.connect(
            lambda: SETTINGS.update_start_in_fullscreen(False)
        )


class RunOnStartUp(YesOrNoSetting):
    def __init__(self, settings_window: SettingsWindow):
        super().__init__(settings_window, "Run on start up")
        appdata_folder = cast(str, os.environ.get("APPDATA"))
        self.lnk_path = os.path.join(
            appdata_folder,
            "Microsoft",
            "Windows",
            "Start Menu",
            "Programs",
            "Startup",
            f"{APP_NAME}.lnk",
        )
        if SETTINGS.run_on_startup:
            self.yes_button.set_picked_status(True)
        else:
            self.no_button.set_picked_status(True)
        self.yes_button.clicked.connect(lambda: SETTINGS.update_run_on_startup(True))
        self.yes_button.clicked.connect(self.make_startup_lnk)
        self.no_button.clicked.connect(lambda: SETTINGS.update_run_on_startup(False))
        self.no_button.clicked.connect(self.remove_startup_lnk)

    def make_startup_lnk(self):
        self.remove_startup_lnk()
        pylnk3_for_file(
            APP_EXE_PATH,
            self.lnk_path,
            MINIMISED_TO_TRAY_ARG,
            "Senpwai startup shortcut",
            work_dir=APP_EXE_ROOT_DIRECTORY,
        )

    def remove_startup_lnk(self):
        if os.path.isfile(self.lnk_path):
            try_deleting(self.lnk_path)


class AllowNotificationsSetting(YesOrNoSetting):
    def __init__(self, settings_window: SettingsWindow):
        super().__init__(settings_window, "Allow notifications uWu?")
        if SETTINGS.allow_notifications:
            self.yes_button.set_picked_status(True)
        else:
            self.no_button.set_picked_status(True)
        self.yes_button.clicked.connect(
            lambda: SETTINGS.update_allow_notifications(True)
        )
        self.no_button.clicked.connect(
            lambda: SETTINGS.update_allow_notifications(False)
        )


class GogoNormOrHlsSetting(SettingWidget):
    def __init__(self, settings_window: SettingsWindow):
        norm_button = GogoNormOrHlsButton(
            settings_window, GOGO_NORM_MODE, settings_window.font_size
        )
        set_minimum_size_policy(norm_button)
        hls_button = GogoNormOrHlsButton(
            settings_window, GOGO_HLS_MODE, settings_window.font_size
        )
        set_minimum_size_policy(hls_button)
        if SETTINGS.gogo_norm_or_hls_mode == GOGO_HLS_MODE:
            hls_button.set_picked_status(True)
        else:
            norm_button.set_picked_status(True)
        norm_button.clicked.connect(lambda: hls_button.set_picked_status(False))
        hls_button.clicked.connect(lambda: norm_button.set_picked_status(False))
        norm_button.clicked.connect(
            lambda: SETTINGS.update_gogo_norm_or_hls_mode(GOGO_NORM_MODE)
        )
        hls_button.clicked.connect(
            lambda: SETTINGS.update_gogo_norm_or_hls_mode(GOGO_HLS_MODE)
        )
        super().__init__(
            settings_window, "Gogo Normal or HLS mode", [norm_button, hls_button]
        )
        self.setting_label.setToolTip(hls_button.toolTip())


class NonZeroNumberInputSetting(SettingWidget):
    def __init__(
        self,
        settings_window: SettingsWindow,
        initial_value: int,
        setting_updater_callback: Callable[[int], None],
        setting_info: str,
        error_on_zero_text: str,
        units: str | None,
        tooltip: str | None = None,
    ):
        self.settings_window = settings_window
        self.setting_updater_callback = setting_updater_callback
        self.number_input = NumberInput(font_size=settings_window.font_size)
        self.number_input.setFixedWidth(60)
        self.number_input.setPlaceholderText(AMOGUS_EASTER_EGG)
        self.number_input.setText(str(initial_value))
        self.input_layout = QHBoxLayout()
        input_widget = QWidget()
        input_widget.setLayout(self.input_layout)
        self.input_layout.addWidget(
            self.number_input, alignment=Qt.AlignmentFlag.AlignLeft
        )
        if units:
            units_label = StyledLabel(None, settings_window.font_size)
            units_label.setText(units)
            set_minimum_size_policy(units_label)
            self.input_layout.addWidget(
                units_label, alignment=Qt.AlignmentFlag.AlignLeft
            )
        else:
            # just to avoid type errors
            units_label = None
        main_layout = QVBoxLayout()
        self.error = ErrorLabel(settings_window.font_size)
        self.error.setText(error_on_zero_text)
        set_minimum_size_policy(self.error)
        self.error.hide()

        main_widget = QWidget()
        main_widget.setLayout(main_layout)
        main_layout.addWidget(self.error)
        main_layout.addWidget(input_widget)
        super().__init__(settings_window, setting_info, [main_widget])
        self.number_input.textChanged.connect(self.text_changed)
        if tooltip:
            self.setting_label.setToolTip(tooltip)
            if units:
                cast(StyledLabel, units_label).setToolTip(tooltip)

    def text_changed(self, text: str):
        if not text.isdigit():
            return
        new_setting = int(text)
        if new_setting == 0:
            self.error.show()
            self.number_input.setText("")
            return
        self.setting_updater_callback(new_setting)


class MaxSimultaneousDownloadsSetting(NonZeroNumberInputSetting):
    def __init__(self, settings_window: SettingsWindow):
        super().__init__(
            settings_window,
            SETTINGS.max_simultaneous_downloads,
            SETTINGS.update_max_simultaneous_downloads,
            "Only allow",
            "Bruh, max simultaneous downloads cannot be zero",
            "simultaneous downloads",
            "The maximum number of downloads allowed to occur at the same time",
        )


class CheckForNewEpsAfterSetting(NonZeroNumberInputSetting):
    def __init__(
        self, settings_window: SettingsWindow, download_window: DownloadWindow
    ):
        self.download_window = download_window
        super().__init__(
            settings_window,
            SETTINGS.check_for_new_eps_after,
            SETTINGS.update_check_for_new_eps_after,
            "Check for new episodes after",
            "Bruh, time intervals can't be zero",
            "hours",
            "Senpwai will check for new episodes of your tracked anime when you start the app\nthen in intervals of the hours you specify so long as it is running",
        )

    def text_changed(self, text: str):
        super().text_changed(text)
        self.download_window.setup_auto_download_timer()
        self.download_window.start_auto_download()


class QualitySetting(SettingWidget):
    def __init__(self, settings_window: SettingsWindow):
        font_size = settings_window.font_size
        self.settings_window = settings_window
        button_1080 = QualityButton(settings_window, Q_1080, font_size)
        button_720 = QualityButton(settings_window, Q_720, font_size)
        button_480 = QualityButton(settings_window, Q_480, font_size)
        button_360 = QualityButton(settings_window, Q_360, font_size)
        self.quality_buttons_list = [button_1080, button_720, button_480, button_360]
        for button in self.quality_buttons_list:
            set_minimum_size_policy(button)
            quality = button.quality
            button.clicked.connect(
                lambda _, quality=quality: self.update_quality(quality)
            )
            if button.quality == SETTINGS.quality:
                button.set_picked_status(True)
        super().__init__(settings_window, "Download quality", self.quality_buttons_list)

    def update_quality(self, quality: str):
        SETTINGS.update_quality(quality)
        for button in self.quality_buttons_list:
            if button.quality != quality:
                button.set_picked_status(False)


class SubDubSetting(SettingWidget):
    def __init__(self, settings_window: SettingsWindow):
        sub_button = SubDubButton(settings_window, SUB, settings_window.font_size)
        set_minimum_size_policy(sub_button)
        dub_button = SubDubButton(settings_window, DUB, settings_window.font_size)
        set_minimum_size_policy(dub_button)
        if SETTINGS.sub_or_dub == SUB:
            sub_button.set_picked_status(True)
        else:
            dub_button.set_picked_status(True)
        sub_button.clicked.connect(lambda: dub_button.set_picked_status(False))
        dub_button.clicked.connect(lambda: sub_button.set_picked_status(False))
        sub_button.clicked.connect(lambda: SETTINGS.update_sub_or_dub(SUB))
        dub_button.clicked.connect(lambda: SETTINGS.update_sub_or_dub(DUB))
        super().__init__(settings_window, "Sub or Dub", [sub_button, dub_button])



#!/usr/bin/env python

import ctypes
import sys

from PyQt6.QtCore import QCoreApplication, Qt
from PyQt6.QtGui import QPalette
from PyQt6.QtWidgets import QApplication
from senpwai.utils.static import APP_NAME, custom_exception_handler
from senpwai.windows.main import MainWindow


def windows_app_initialisation():
    # Change App ID to ensure task bar icon is Swnpwai's icon instead of Python for pip installs
    # StackOverflow Answer link: https://stackoverflow.com/questions/1551605/how-to-set-applications-taskbar-icon-in-windows-7/1552105#1552105
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(APP_NAME)
    # Convert the app title to a null-terminated C string
    app_title_c = (APP_NAME + "\0").encode("UTF-8")
    # Check if the app is already running by searching for its window
    window = ctypes.windll.user32.FindWindowA(None, app_title_c)
    if window != 0:
        # If the app is running, bring it into focus
        ctypes.windll.user32.ShowWindow(window, 9)  # 9 is SW_RESTORE
        ctypes.windll.user32.SetForegroundWindow(window)
        sys.exit(0)


def main():
    if sys.platform == "win32":
        windows_app_initialisation()

    QCoreApplication.setApplicationName(APP_NAME)
    args = sys.argv
    app = QApplication(args)
    app.setApplicationName(APP_NAME)
    palette = app.palette()
    palette.setColor(QPalette.ColorRole.WindowText, Qt.GlobalColor.white)
    app.setPalette(palette)

    window = MainWindow(app)
    app.setWindowIcon(window.senpwai_icon)
    window.show_with_settings(args)
    sys.excepthook = custom_exception_handler
    sys.exit(app.exec())


if __name__ == "__main__":
    main()
